/*
** Module   : MenuWidget.java
** Abstract : server side menu widget implementation
**
** Copyright (c) 2015-2020, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 VIG 20141029 Created initial version.
** 002 VIG 20150115 Changed base class, added pushScreenDefinition calls, 
**                  added initialization code.
** 003 VIG 20150501 Implemented CommonHandleTree interface.
** 004 VIG 20150823 Fixed firstChild and lastChild methods.
** 005 HC  20170131 Added runtime support for dynamic MENU* widgets.
** 006 CA  20170929 Fixed MENU:POPUP-ONLY attribute getter return type.
**                  Improved menu registration with client-side.
** 007 CA  20180130 Replaced pushScreenDefinition() with pushWidgetAttr() for cases when business 
**                  logic is assigning a widget attribute.
** 008 HC  20180523 Fixed MENU resource leak, static MENU was not deregistered due to apparently 
**                  incorrect interpretation of its visibility flag.
** 009 CA  20180526 Fixed SENSITIVE setter.
** 010 CA  20180710 Fixed menu delete.
** 011 CA  20190322 Avoid unnecessary client-side trips if the attribute isn't changing on set.
** 012 CA  20200207 Fixed a regression related to ProcedureManager H057 changes related to static
**                  resource delete.
** 013 HC  20200313 Javadoc fixes.
**     GES 20200306 Shifted to TitledElement.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.util.*;

/**
 * This class represents menu widget.
 */
@LegacyResource(resource = LegacyResource.MENU)
public class MenuWidget
extends MenuContainerWidget<MenuConfig>
implements MenuInterface,
           TitledElement
{
   /** OWNER property: window for menubar or other widget for popup menu */
   private GenericWidget<?> owner;
   
   /**
    * Default constructor.  
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public MenuWidget(boolean dynamic)
   {
      super(dynamic, new MenuConfig());
      config.popupOnly = !dynamic;
   }

   /**  
    * POPUP-ONLY attribute (Graphical interfaces only)
    *
    * @return  The attribute value.
    */
   public logical isPopupOnly()
   {
      return new logical(config.popupOnly);
   }

   /**
    * Set POPUP-ONLY option.
    * 
    * @param   popupOnly
    *          <code>true</code> if this menu is popup-only, <code>false</code> otherwise.
    */
   public void setPopupOnly(boolean popupOnly)
   {
      if (config.popupOnly == popupOnly)
      {
         return;
      }
      
      config.popupOnly = popupOnly;
      pushWidgetAttr("popupOnly", config.popupOnly);
   }

   /**
    * Set POPUP-ONLY option.
    * 
    * @param   popupOnly
    *          <code>true</code> if this menu is popup-only, <code>false</code> otherwise.
    */
   public void setPopupOnly(logical popupOnly)
   {
      if (popupOnly == null || popupOnly.isUnknown())
      {
         String msg = "**Unable to assign UNKNOWN value to attribute POPUP-ONLY on MENU " + name();
         ErrorManager.recordOrShowError(4083, msg, false, false, false);
      }
      else 
      {
         setPopupOnly(popupOnly.booleanValue());
      }
   }
   
   /**
    * Get handle of an owner widget.
    * 
    * @return  See above.
    */
   public handle getOwnerHandle()
   {
      return owner == null ? new handle() : owner.asWidgetHandle();
   }

   /**
    * Get owner widget.
    * 
    * @return  See above.
    */
   public GenericWidget<?> getOwner()
   {
      return owner;
   }

   /**
    * Set owner widget.
    * 
    * @param   owner
    *          The widget that will be own this menu.
    */
   public void setOwner(GenericWidget<?> owner)
   {
      if (this.owner == owner)
      {
         return;
      }

      this.owner = owner;
      this.config.ownerId = owner == null ? 0 : owner.getId();
      pushScreenDefinition();
   }
   
   /**
    * Sets the color for the menu title.
    *
    * @param    titleColor
    *           The overridden color for the title.
    */
   public void setTitleColor(ColorSpec titleColor)
   {
      config.titleDColor = titleColor.convert();
      pushWidgetAttr("titleDColor", config.titleDColor);
   }
   
   /**
    * Finds menu-item in this menu tree. It can appear on different levels.
    * Menu-item is searched using some sort of depth-first algorithm.
    * See {@link #findMenuItemInt(String)}
    * 
    * @param   name
    *          MENU-ITEM name.
    *          
    * @return  concrete {@link MenuItemWidget} or null if not found.
    */
   public MenuItemWidget findMenuItem(String name)
   {
      return findMenuItemInt(name);
   }

   /**
    * Finds sub-menu in this menu tree. It can appear only once. 
    * Simple depth-first search is used.
    * 
    * @param   name
    *          Sub-menu name.
    * 
    * @return  concrete {@link SubMenuWidget} or null if not found.
    */
   public SubMenuWidget findSubMenu(String name)
   {
      return findSubMenuInt(name);
   }
   
   /**
    * Accesses the first window in the list of contained windows, having this window as parent.
    *
    * @return   The first widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle firstChild()
   {
      Iterator<GenericWidget<?>> it = widgets().iterator();
      return it.hasNext() ? new handle(it.next()) : new handle();
   }
   
   /**
    * Accesses the last window in the list of contained windows, having this window as parent.
    *
    * @return   The last widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle lastChild()
   {
      Iterator<GenericWidget<?>> it = widgets().iterator();
      GenericWidget<?> last = null;
      while (it.hasNext())
      {
         last = it.next();
      }
      return last != null ? new handle(last) : new handle();
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(boolean s)
   {
      if (config.enabled == s)
      {
         return;
      }

      // changing the SENSITIVE attribute does not affect the widget's visibility
      config.enabled = s;

      pushWidgetAttr("enabled", config.enabled);
   }

   /**
    * Detach the specified descendant from this menu.
    * 
    * @param    comp
    *           The descendant to detach.
    */
   void detachMenu(MenuItemWidget comp)
   {
      if (isBatch())
      {
         return;
      }
      
      batch(true);
      
      // remove from parent first
      GenericWidget<?> parent = (GenericWidget<?>) comp.getParent();
      if (parent != null)
      {
         ((MenuContainerWidget) parent).removeWidget(comp);
      }

      LogicalTerminal.destroyWidgets(new int[] { comp.getId() } );

      batch(false, true);
   }

   /**
    * Detach the specified descendant from this menu.
    * 
    * @param    comp
    *           The descendant to detach.
    */
   void detachMenu(MenuContainerWidget<?> comp)
   {
      if (isBatch())
      {
         return;
      }
      
      batch(true);
      
      // remove from parent first
      GenericWidget<?> parent = comp instanceof MenuWidget 
                                   ? ((MenuWidget) comp).getOwner()
                                   : (GenericWidget<?>) comp.getParent();
      if (parent != null)
      {
         if (comp instanceof MenuWidget)
         {
            parent.detachMenu((MenuWidget) comp);
         }
         else
         {
            ((MenuContainerWidget) parent).removeWidget(comp);
         }
      }

      // get the sub-tree and delete each component
      List<GenericWidget<?>> menuTree = new LinkedList<>();
      comp.listMenuTree(menuTree);
      int[] menuTreeIds = new int[menuTree.size()];
      
      int i = 0;
      for (GenericWidget<?> descendant : menuTree)
      {
         if (descendant != comp)
         {
            ProcedureManager.deleteResource(null, descendant);
         }
         
         menuTreeIds[i++] = descendant.getId();
      }
      
      LogicalTerminal.destroyWidgets(menuTreeIds);
      
      batch(false, true);
   }

   /**
    * Creates new menu as a proxy object (as a local resource).  This loads
    * the static menu definition.
    *
    * @param    <T> 
    *           Menu definition class type.
    * @param    defCls
    *           The class of menu definition.
    * @param    name
    *           The name of menu.          
    *
    * @return   The created menu widget.
    */
   public static <T extends MenuDefinition> T createStaticMenu(Class<T> defCls, String name)
   {
      MenuWidget menu = new MenuWidget(false);
      menu.name(name);
      
      final T menuDef;
      try
      {
         menuDef = defCls.getConstructor(new Class<?>[]{}).newInstance();
         menu.config.id = new WidgetId(LogicalTerminal.registerMenu(menu, true));
         menu.batch(true);
         menuDef.setSelf(menu);
         menuDef.setup(menu);
         menu.coreInitialize(defCls, menuDef, false);
         menu.batch(false, true);
      }
      catch (NoSuchMethodException    | 
             SecurityException        | 
             InstantiationException   | 
             IllegalAccessException   | 
             IllegalArgumentException | 
             InvocationTargetException e)
      {
         throw new RuntimeException("Failed to instantiate menu definition class", e);
      }
      
      return menuDef;
   }

   /**
    * Creates new shared menu.
    *
    * @param    <T> 
    *           Menu definition class type.
    * @param    defCls
    *           The class of menu definition.
    * @param    name
    *           The name of menu.          
    *
    * @return   The created menu definition.
    */
   public static <T extends MenuDefinition> T createStaticSharedMenu(Class<T> defCls, String name)
   {
      final T menuDef = createStaticMenu(defCls, name);
      
      menuDef.getSelf().setShared(true);
      
      // export the menu. All shared menus are global.
      SharedVariableManager.addMenu(name, menuDef);

      return menuDef;
   }

   /**
    * Import an existing menu
    * 
    * @param    <T> 
    *           MenuDefinition interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The menu name.
    *
    * @return   The menu frame.
    */
   @SuppressWarnings("unchecked")
   public static <T extends MenuDefinition> T importStaticSharedMenu(Class<T> iface, String name)
   {
      MenuDefinition oldMenu = (MenuDefinition) SharedVariableManager.lookupMenu(name);
      
      return (T) oldMenu;
   }
}
