/*
** Module   : BrowseWidget.java
** Abstract : server side BROWSE widget class
**
** Copyright (c) 2005-2020, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23481 Created initial version.
** 002 SIY 20060307   @25125 Major update for complete BROWSE support.
** 003 SIY 20060319   @25125 Fixed update.
** 004 SIY 20060328   @25276 Fixed reposition code, added a number of 
**                           methods to set various flags and attributes.
** 004 SIY 20060406   @25400 Added setCentered() methods. 
** 005 SIY 20060427   @25816 Added another signature for  setRepositionerRow() method.
** 006 ECF 20060501   @25917 Change to getRowCount. PreselectQueryWrapper
**                           is now obsolete; use new P2JQuery.isPreselect method instead.
** 007 ECF 20060618   @27424 Change to setQuery(). Ensure new query is set to be scrolling.
** 008 GES 20060803   @28366 Some cleanup and refactoring.
** 009 GES 20060804   @28430 Honor new iterate() processing which only
**                           happens at the top of the block body.
** 010 GES 20060808   @28490 Use int instead of long for the backing 
**                           methods that support the server exports.
** 011 ECF 20061019   @30532 Changed reposition(). Removed adjustment code for query off-end.
** 012 ECF 20061023   @30610 Made BrowseWidget tolerant of a query which returns no results. When
**                           a query is set into the widget, its errorIfNull property is set to
**                           false. QueryOffEndException is caught and ignored in finishSetup()
**                           when trying to retrieve the first record.
** 013 ECF 20061103   @30963 Implement RepositionListener interface. This is necessary to receive
**                           notifications of reposition events from the associated query.
**                           Made additional changes to support changes in the P2JQuery interface
**                           and to support the assumption that a new record is fetched when
**                           the associated query is repositioned.
** 014 SIY 20061105   @30983 Finished support for notifications and added getCurrentRow() method.
** 015 GES 20061110   @31077 Fixed abend in getCurrentRow().
** 016 SIY 20061114   @31360 Fixed some repositioning issues. Removed initial query positioning
**                           which did break  browse behavior (initially selected row is
**                           the current row in query).
** 017 ECF 20061129   @31479 Fixed getRowCount(). This was aborting with a QueryOffEndException
**                           in the case of a query which had no results. This exception is now
**                           caught and handled.
** 018 ECF 20061204   @31578 Use new P2JQuery.setBrowsed() method. Also, getCurrentRow() reports
**                           -1 instead of 0 for current row if query is off-end, since 0 is a
**                           valid row number.
** 019 SIY 20070316   @32421 Set column ordinal when it is added to browse.
** 020 SIY 20070507   @33425 Changed handling of record set size request and other 
**                           relevant changes.
** 021 SIY 20070515   @33755 Fixed positioning. 
** 022 SIY 20070605   @33963 Fixed restoring of the position after getRows.
** 023 NVS 20070618   @34083 Fixed ignoreRepo state variable rebalancing.
**                           Due to exceptions in finally{} clauses it could have been left
**                           unbalanced at times, causing legitimate requests to be ignored.
** 024 NVS 20070622   @34206 Fixed regression introduced by #022. reposition() calls throw
**                           QueryOffEndException when the result set is empty in getRows().
** 025 ECF 20070928   @35274 Fixed setCurrentRow(). Was repositioning to
**                           first row after running off-end at last row.
** 026 ECF 20071011   @35446 Further adjustment to setCurrentRow(). Avoid
**                           multiple repositions in case of selecting last browse row.
** 027 ECF 20080208   @37029 Fixed fetchSelectedRow(). This method was  improperly using its
**                           argumentas the row index, rather than as an index within the
**                           list of selected rows.
** 028 GES 20080320   @37581 Added TYPE attribute.
** 029 ECF 20080509   @38246 Added safety code to getRows(). We must first
**                           check if any rows are available before trying to fetch them.
** 030 SVL 20080512   @38277 setScrolling() and setErrorIfNull(false) were
**                           removed from setQuery(), reposition(1) was added to setQuery().
** 031 SVL 20080604   @38536 currentRowImpl() is used instead of currentRow().
** 032 SIY 20080625   @38923 Minor fixes in javadoc and removed unnecessary type cast.
** 033 CA  20080925   @39961 Support for dynamic title.
** 034 ECF 20081106   @40318 Improved getRows(). Use a dynamic list instead of a fixed array,
**                           in case the row count changes mid-query. Break fetch loop on
**                           query off-end condition.
** 035 CA  20081203   @40813 getColumns() must set the ID of the browse column, before sending
**                           it to the client.
** 036 CA  20081210   @40830 Events triggered via APPLY will force the buffer to load the
**                           currently selected row.
** 037 CA  20081210   @40842 Implemented CLOSE QUERY statement. On query close, the browse will
**                           reset its data.
** 038 CA  20081212   @40886 Fixed a regression in setCurrentRow, caused by
**                           #036: rowIndex must always be incremented (so it will be 1-based).
** 039 SIY 20090213   @41306 Added 'force' parameter into refreshStartRow signature.
**                           Added forceRefresh() method.
** 040 SIY 20090706   @43055 Removed apply(). Changes in query behavior made it obsolete.
** 041 CA  20090906   @43831 Fixed FETCH-SELECTED-ROW() - an error message  must be shown if
**                           the received index is unknown. Fixed SET-REPOSITIONED-ROW() - if
**                           the received row number is unknown, it must return false.
** 042 SVL 20091030   @44295 Offend state is preserved while execution of the  getRowCount().
** 043 SVL 20091105   @44316 Into getRowCount() set the query cursor final state to "not between
**                           rows" (if the query allows fetch on reposition).
** 044 SIY 20101122          Minor cleanups.
** 045 SIY 20111101          Updated imports.
** 046 SIY 20111202          Updated imports.
** 047 CA  20121022          Added support for optional query and column list 
**                           clauses.
** 048 CA  20130221          Refactored usage of TYPE attribute.
** 049 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 050 OM  20130320          Added implementation for ResultListHandler interface.
** 051 VMN 20130411          Added boolean parameter modal to recordOrShowError() method.
** 052 OM  20130527          Added int64 support for accessors.
** 053 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 054 CA  20131024          Only QUERY resources allow ResultListHandler API calls - so cast the
**                           query to the ResultListHandler type before calling.
** 055 CA  20131115          Changes related to properly instantiate dynamic resources.
** 056 GES 20140627          Changed color attribute method names to match naming standards.
** 057 CA  20140711          Enhanced font support, for GUI clients.
** 058 VIG 20140810          Added DESELECT-ROWS error processing.
** 059 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 060 SVL 20140909          In getRowCount use query.isNativelyPreselect rather than
**                           query.isPreselect in order to correctly detect adaptive queries as
**                           dynamic.
** 061 CA  20141104          Removed set/getWidth: the BaseEntity's set/getWidthChars APIs need to
**                           be emitted/used.
** 062 SVL 20141102          Added some attributes and functions.
** 063 SVL 20141209          Implemented some functions. Updated error messages in
**                           setNumLockedColumns.
** 063 GES 20141124          Added modified and query attributes.
** 064 SVL 20150212          Modified getRows and getCurrentRow.
** 065 SVL 20150305          Added refresh mode for DELETE-CURRENT-ROW. Fixed case of
**                           unavailable records in getRows.
** 066 SVL 20150331          Implemented DELETE-* functions.
** 067 SVL 20150411          Implemented SELECT-{NEXT|PREV}-ROW, ROW-MARKERS, MAX-DATA-GUESS.
** 068 HC  20150323          Removed business logic from config classes.
** 069 SVL 20150420          Implemented REFRESHABLE.
** 070 SVL 20150508          Implemented insertRow. Implemented "fill-in per row" approach.
** 071 SVL 20150625          Added getRowCount(). Fixed reverse fetching in getRows().
** 072 CA  20150722          The BROWSE transfers to the client-side only the IDs of its 
**                           associated BrowseColumn widgets.  The actual column definitions are
**                           transferred via pushScreenDefinition.
** 073 SVL 20150722          Implemented MODIFIED attribute, SET-REPOSITIONED-ROW.
** 074 SVL 20150804          Reflect changes in Down, Multiple and QueryAssociable interfaces.
** 075 SVL 20150821          Added support for dynamic browses.
** 076 GES 20150923          Javadoc fix.
** 077 SVL 20150922          Added some colors and fonts accessors. In cleanup() pushing screen
**                           definitions for columns is deferred until browse is deleted.
** 078 OM  20150922          Overloaded attribute declarations with missing long/int64 signatures.
**                           Added NUM-ITERATIONS, NO-VALIDATE, BUFFER-FIELD stubs.
** 079 SVL 20151019          Added COLUMN-SCROLLING attribute.
** 080 IAS 20151028          Added NO-VALIDATE attribute.
** 081 SVL 20151027          Implemented visibleIterations().
**     SVL 20151029          Use queryRepositioned instead of refreshScrollRow.
** 082 SVL 20151129          Browse now implements Labels interface.
** 083 SVL 20151230          Store widget name in config (for error messages).
** 084 EVL 20160217          Clean up comments from symbols invalid for Solaris to compile.
** 085 IAS 20160229          LOAD-MOUSE-POINTER support.
** 086 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 087 SVL 20160303          Do not set fill-in widgets widths initially. That may cause
**                           "Attribute WIDTH must be greater than zero on FILL-IN widget" error
**                           message. Added moveColumn.
** 088 SVL 20160419          Implemented getMinHeightChars.
** 089 SVL 20160608          Fixed displaying of queries with OUTER join.
** 090 HC  20160714          Fixed the data type of the converted GET-BROWSE-COLUMN in the Java
**                           APIs.
** 091 HC  20160716          Implemented runtime support for BROWSE-COLUMN:RESIZABLE attribute.
** 092 SVL 20160811          Support for reassigned browsed queries with another backing buffers.
** 093 SBI 20160921          Changed to override setFrame(GenericFrame frame) to add this browse
**                           widget to its frame's field group, changed setQuery(.) to fix
**                           QueryOffEndException on query.reposition(1) if this query is empty,
**                           fixed a scenario in which finishSetup() is done, but the query hasn't
**                           been set yet, changed the type of a browse column widget created by
**                           addLikeColumn is dynamic.
** 094 SVL 20160922          Code refactoring because of switching from BrowseConfig.queryId to
**                           BrowseConfig.queryAssigned.
** 095 SBI 20161004          Fixed addLikeColumn to set the column data format.
** 096 EVL 20161021          Preventing from applying empty format string for column widget.
** 097 SVL 20161029          config.queryAssigned is properly set only on query assignment.
**                           getRows() now returns null if the query is closed.
** 098 CA  20161123          If a BROWSE has not-available record buffers (from an outer-join),
**                           then do not show an error, just assume unknown value for the fields
**                           in these buffers.
** 099 SVL 20170126          Added getCurrentColumn.
**     GES 20170130          Override the SCREEN-VALUE attribute to notify that it is not
**                           settable or queryable.
** 100 SVL 20170417          Implemented search by key.
** 101 EVL 20170529          Adding Droppable interface to handle file list dropping.  Adding
**                           ROW-RESIZABLE attribute support.
** 102 SVL 20170809          ROW-DISPLAY events are triggered at server side. Added
**                           ADD-CALC-COLUMN, ROW-HEIGHT-*, NO-EMPTY-SPACE.
** 103 CA  20170827          Added stub for NUM-ENTRIES attribute getter (undocumented).
**     SVL 20170829          Added an overloaded version of addCalcColumn.
**     SVL 20170831          Support for toggle-box columns.
** 104 SVL 20171025          Fixed setReadOnly. Added setNoValidateOption. Support for
**                           AUTO-VALIDATE. Support for validation statements in dynamic columns.
** 105 SVL 20171116          Refactored so error 6422 can be rised on the client side too.
** 106 OM  20171122          Used WriteProtectable interface for READ-ONLY attribute.
**     SBI 20171124          Removed setFrame in order to delegate to GenericWidget.setFrame.
** 107 GES 20171220          Added a TODO comment.
** 108 OM  20171129          Release the buffers when browse is positioned on a virtual record.
**     CA  20171202          The cell editor's name is inherited from the column.
**     SVL 20171204          Added refreshCurrentRow.
**     CA  20171204          finishSetup() is responsible for createEditors().
** 109 CA  20180130          Replaced pushScreenDefinition() with pushWidgetAttr() for cases when
**                           business logic is assigning a widget attribute.
** 110 SVL 20171225          Fixed error handling in fetchSelectedRow.
** 111 SVL 20180126          ALLOW-COLUMN-SEARCHING is true by default for dynamic browses.
** 112 SVL 20180209          Added REPORT attribute. Changes related to enhanced browse and
**                           column:VISIBLE attribute.
** 113 SVL 20180321          Data for invisible columns should be preserved on the client side
**                           too.
** 114 EVL 20180330          Fix for ability to set size in GUI mode.
** 115 HC  20180410          Renamed ReportOutputFormat to MediaType.
** 116 EVL 20180423          Renamed Report to FwdReport.
** 117 CA  20180527          ROW-DISPLAY trigger may change the cell's value (i.e. record's field)
**                           for non-computed columns, too.
** 118 SVL 20180529          Implemented enhanced configuration management.
**     OM  20180620          Added support for dynamically configured filtering and sort criteria.
**                           Made column key names unique.
** 119 SVL 20180901          Implemented dynamic sorting/filtering. Improved toString().
** 120 SVL 20180913          Support for labelsHeightNoEnh. Row and column separators can be
**                           switched independently.
**     OM  20180924          Filtering on decimal fields uses STRING(f, format) function.
**     OM  20181003          Only exports to PDF format will take into consideration pagination.
**     SVL 20181004          Added getVisibleColumns.
** 121 OM  20181015          Moved IS_IGNORE_PAGINATION parameter into design template.
**     SVL 20181007          Added alternating background color.
**     SVL 20181008          Added applyEnhancedLayout, changeEnhancedLayout and exporting.
**     OM  20181017          Small optimization: avoid multiple lookup of context value.
** 122 SVL 20190108          Fixed setNoColumnScrolling.
** 123 SBI 20190208          Changed getRows to trigger row-display and then to evaluate its
**                           column value if the screen value is not set.
**     EVL 20190218          Fix for issues when need to refresh current row while it is already
**                           under changing the column screen values.
**     SBI 20190301          Changed addCalcColumn according to DisplayFormat.instanceOfType,
**                           changed to mark the column screen values invalid if the associated
**                           query is closed. 
** 124 SVL 20190303          Fixed extra error message related to NUM-LOCKED-COLUMNS.
**                           Fixed bug related to values set in ROW-DISPLAY trigger.
**     SVL 20190321          Re-worked ROW-DISPLAY logic: SCREEN-VALUE set in a ROW-DISPLAY
**                           trigger is converted immediately rather than postponed until
**                           rendering time.
**     SVL 20190325          Updated signature of notifyColumnsUpdated.
** 125 EVL 20190319          Adding (MIN|MAX)-(HEIGHT|WIDTH)-CHARS attributes support.
** 126 EVL 20190319          Adding (MIN|MAX)-HEIGHT-CHARS attributes support.  However only
**                           MIN-HEIGHT-CHARS is available to use in read-only mode.  Addidng
**                           NUM-VISIBLE-COLUMNS attribute support. Added method ADD-COLUMNS-FROM,
**                           runtime implementation completed.
**                           query is closed.
** 127 SVL 20190322          Updated createResultListEntry().
** 128 CA  20190423          Added more addCalcColumn overloaded versions.
** 129 EVL 20190325          More fix for issue when column is in editing mode and we need to
**                           prefer screen value over variable one to consider as column current
**                           value.
**     SBI 20190403          Changed queryClosed() to cleanup dirty flags after the client has
**                           been notified.
** 130 SVL 20190412          Added support for combo-box columns.
** 131 SVL 20190614          Javadoc update.
** 132 CA  20190703          Added addCalcColumn(character, character, character, String) version.
** 133 SBI 20190517          Added releaseBuffers() to stress this unit of functionality.
**     SBI 20190623          Removed dirty (explicitScreenValueAssigned) flag
** 134 EVL 20190815          Fix for reposition(1) issue with not yet populated query.  Another
**                           version for this fix, guarding from more exception instead.
**     HC  20190823          When query is assigned to a browse, the existing browse columns must
**                           be cleared. When query is deleted it must be unassigned from browse.
**     EVL 20190823          More rework for reposition issue, making it postponed some times.
** 135 SVL 20191002          Properly implemented create/deleteResultListEntry.
** 136 HC  20191019          Improved the logic preventing recursive triggers to consider the
**                           cases when the widget matching the trigger is different from the
**                           source widget.
** 137 AIL 20191105          Fixed issue when columns weren't cleared while query was assigned.
**     SVL 20191111          Fixed checkNoRecords for the case when a browse has a single row
**                           which is new.
** 138 EVL 20191122          Adding COMBO-BOX widget as column data type support.  Looks like not
**                           a complete implementation(must be separately tested), just avoid the
**                           unwanted error messages generation.  Only CHAR or LONGCHAR data types
**                           are now supported.
** 139 GES 20200213          Implemented NUM-ENTRIES attribute support.
** 140 RFB 20200319          Modified implementation of NUM-ENTRIES to get the value from the client.
**     GES 20200220          Added TitledElement.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.reporting.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.*;

import static com.goldencode.p2j.ui.BrowseColumnWidgetType.*;
import static com.goldencode.p2j.ui.RowDisplayStrategy.*;
import static com.goldencode.p2j.ui.client.gui.BrowseGuiImpl.MIN_BROWSE_ROWS;
import static com.goldencode.p2j.persist.P2JQuery.*;

/**
 * Server side representation of the Browse widget.
 */
@LegacyResource(resource = LegacyResource.BROWSE)
public class BrowseWidget
extends BaseEntity<BrowseConfig>
implements BrowseElement,
           BrowseInterface,
           QueryAssociable,
           RepositionListener,
           ResultListHandler,
           ScrollbarVerticalElement,
           Multiple,
           Down,
           Labels,
           Hoverable,
           Droppable,
           RowDisplayDrawListener,
           WriteProtectable,
           TitledElement,
           QueryOpenedListener
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(BrowseWidget.class.getName());

   /** Error text for an error message with code 6422. */
   public static String ERROR_TEXT_6422 =
         "The %s specified for BROWSE BROWSE %s " + // double BROWSE is correct
               "is invalid. The change to this attribute forces a change in the " +
               "number of viewport rows below the minimum number of required rows (%s)";

   /** Default value for the MAX-DATA-GUESS attribute. */
   private static int DATA_GUESS_DEFAULT_VALUE = 100;
   
   /** Underlying query.  May be null. */
   private QueryWrapper query = null;
   
   /** Underlying query.  May be null. */
   private QueryWrapper originalQuery = null;
   
   /** The registered storage for the current browse title. */
   private CharacterExpression dynTitle = null;
   
   /** List of columns in browse. */
   private ArrayList<BrowseColumnWidget> columns = new ArrayList<>();

   /** Widget initialization flag. */
   private boolean initDone = false;

   /** Default 4GL browse state: matches non-enhanced view. */
   private EnhancedBrowseConfig default4glState = null;

   /** <code>true</code> if initial handling of enhanced configuration is done. */
   private boolean enhancedInitDone = false;
   
   /** Flag to ignore browse-driven reposition events */
   private int ignoreRepo = 0;

   /** Cached value of client-side (real) DOWN value. */
   private Integer pageRowsCount = null;

   /**
    * Maximum known row index (1-based). <code>0</code> if there are no rows. <code>null</code>
    * if we haven't tried to fetch any rows with the underlying query yet.
    */
   private Integer maximumKnownRowIndex = null;

   /**
    * <code>true</code> if an actual number of fetched rows is used as MAX-DATA-GUESS.
    * <code>false</code> if the default value is used.
    */
   private boolean realDataGuessUsed = false;

   /**
    * Pending query reopening state while REFRESHABLE is off.
    */
   private PendingOpenState pendingReopen = PendingOpenState.NO_REOPEN;

   /**
    * Pending 0-based repositioned row index (i.e. the last row the query was repositioned to)
    * while REFRESHABLE is off. <code>-1</code> if there is no repositioned row.
    */
   private int pendingRepositionRow = -1;

   /**
    * Row for which ROW-DISPLAY event is being triggered. <code>null</code> if we are not in
    * ROW-DISPLAY event.
    */
   private BrowseRow rowInDisplay = null;

   /**
    * <p>While we are getting new rows, this is the list of the rows that were already fetched and
    * can be drawn if a pause takes place during ROW-DISPLAY (e.g. caused by PAUSE statement or a
    * message box). See {@link LogicalTerminal#rowDisplayDrawListener} for the full list of
    * listeners that can trigger pause action and generation of "intermediate" rows.</p>
    * <p>Actual content of the array can differ depending on what type of rows iteration is
    * performed.</p>
    * <p><code>null</code> if we are not in
    * {@link #getRows(char, int, int, int, int, ScreenBuffer[])}.</p>
    */
   private ArrayList<BrowseRow> intermediateRows = null;

   /** Handle to the browse {@link FwdReport} object. */
   private handle report;

   /** Counter for generating unique keys for calc columns (for enhanced browse). */
   private int calcColumnCounter = 0;

   /** Set of column keys. Used by enhanced browse to check for duplicate column keys. */
   private Set<String> ehColumnKeys = null;
   
   /** 
    * Keeping track of the columns already added to browse for given table.  Only column added
    * by ADD-COLUMNS-FROM() method is considering here.
    */
   private Map<String, ArrayList<String>> colsAddedFrom = null;
   
   /** Flag indicating we need to defer query reposition until it become opened. */
   private boolean deferQueryReposition = false;

   /**
    * Default constructor.
    */
   public BrowseWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public BrowseWidget(boolean dynamic)
   { 
      super(dynamic, new BrowseConfig());

      if (dynamic)
      {
         config.columnScrolling = false;
         config.allowColumnSearching = true;
      }
      
      EnhancedBrowseConfigManager manager = EnhancedBrowseConfigManager.getInstance();
      config.applyEnhancedLayout  = manager.isLayoutApplied();
      config.changeEnhancedLayout = manager.canChangeLayout();
      config.exporting            = manager.isExporting();
      setEnhancedSorting(true);   // manager.isEnhancedSorting() is applied
      setEnhancedFiltering(true); // manager.isEnhancedFiltering() is applied
   }

   /**
    * Get the number of entries visible in the browse.  This {@code NUM-ENTRIES} attribute is
    * undocumented but does work in the 4GL. It returns the number of visible rows with data,
    * which means it varies between 0 and the DOWN value. This value comes from the client.
    * 
    * @return   The number of visible rows with data.
    */
   public integer getNumEntries()
   {
      return new integer(LogicalTerminal.getBrowseNumEntries(this));
   }
  
   /**
    * {@code NUM-ENTRIES} is a read-only attribute, calls to this method will generate an error.
    * 
    * @param    numEntries
    *           The new attribute value. 
    */
   public void setNumEntries(int64 numEntries)
   {
      readOnlyError("NUM-ENTRIES");
   }
  
  /**
    * {@code NUM-ENTRIES} is a read-only attribute, calls to this method will generate an error.
    * 
    * @param    numEntries
    *           The new attribute value. 
   */
   public void setNumEntries(long numEntries)
   {
      readOnlyError("NUM-ENTRIES");
   }
   
   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   @Override
   public logical isReadOnly() 
   {
      return new logical(config.readOnly);
   }
   
   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   @Override
   public void setReadOnly(logical r) 
   {
      if (config.readOnly == r.booleanValue())
      {
         return;
      }
      
      config.readOnly = r.booleanValue();
      config.editPossible = calcEditPossible();
      if (!config.readOnly)
      {
         createEditors();
      }

      pushScreenDefinition();
   }
   
   /**
    * Get the HELP writable attribute.
    * 
    * @return   The current value of the HELP attribute.
    */
   public character getHelp()
   {
      String help = config.help;
      return (help == null) ? new character() : new character(help);
   }

   /**
    * Queries if the SCREEN-VALUE attribute for this widget was changed.
    * 
    * @return  If <code>true</code>, the widget's screen value was changed.
    */
   public logical isModified()
   {
      return new logical(config.modified);
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(final boolean modified)
   {
      if (config.modified == modified)
      {
         return;
      }
      
      ConfigHelper.setModified(this.config, modified);

      pushWidgetAttr(new String[] { "modified", "state"}, 
                     new Object[] { config.modified, config.state});
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(logical modified)
   {
      if (modified == null || modified.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute MODIFIED on BROWSE " + widgetName(),
               false);

         return;
      }
      
      setModified(modified.booleanValue()); 
   }
   
   /**
    * Get the MULTIPLE attribute.
    * 
    * @return   The value of the MULTIPLE attribute.
    */
   public logical isMultiple()
   {
      return new logical(config.multiple);
   }

   /**
    * Set value of the MULTIPLE attribute.
    * 
    * @param    multiple
    *           New value of the attribute.
    */
   public void setMultiple(boolean multiple)
   {
      if (config.realized)
      {
         errorAlreadyRealized("MULTIPLE");
         return;
      }

      if (config.multiple == multiple)
      {
         return;
      }
      
      config.multiple = multiple;
      pushWidgetAttr("multiple", config.multiple);
   }
   
   /**
    * Set value of the MULTIPLE attribute.
    * 
    * @param    multiple
    *           New value of the attribute.
    */
   public void setMultiple(logical multiple)
   {
      if (multiple.isUnknown())
         return;
      
      setMultiple(multiple.booleanValue());
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param    flag
    *           New value of the attribute.
    */
   public void setScrollbarVertical(boolean flag)
   {
      if (config.scrollbarVertical == flag)
      {
         return;
      }
      
      config.scrollbarVertical = flag;
      pushWidgetAttr("scrollbarVertical", config.scrollbarVertical);
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param    flag
    *           New value of the attribute.
    */
   public void setScrollbarVertical(logical flag)
   {
      if (flag == null || flag.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-VERTICAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarVertical(flag.booleanValue());
   }

   /**
    * Get value of the SCROLLBAR-VERTICAL attribute.
    *
    * @return  value of the attribute.
    */
   public logical isScrollbarVertical()
   {
      return new logical(config.scrollbarVertical);
   }
   
   /**
    * Set OVERLAY attribute.
    * 
    * @param    overlay
    *           New attribute value.
    */
   public void setOverlay(boolean overlay)
   {
      if (config.overlay == overlay)
      {
         return;
      }
      
      config.overlay = overlay;
      pushWidgetAttr("overlay", config.overlay);
   }
   
   /**
    * Set OVERLAY attribute.
    * 
    * @param    overlay
    *           New attribute value.
    */
   public void setOverlay(logical overlay)
   {
      if (overlay.isUnknown())
         return;
      
      setOverlay(overlay.booleanValue());
   }

   /**
    * Set value of the ROW-MARKERS attribute.
    * 
    * @param    markers
    *           New value of the attribute.
    */
   public void setRowMarkers(boolean markers)
   {
      if (config.realized)
      {
         errorAlreadyRealized("ROW-MARKERS");
         return;
      }

      if (config.rowMarkers != null && config.rowMarkers == markers)
      {
         return;
      }
      
      config.rowMarkers = markers;
      pushWidgetAttr("rowMarkers", config.rowMarkers);
   }
   
   /**
    * Set value of the ROW-MARKERS attribute.
    * 
    * @param    markers
    *           New value of the attribute.
    */
   public void setRowMarkers(logical markers)
   {
      if (markers.isUnknown())
         return;
      
      setRowMarkers(markers.booleanValue());
   }

   /**
    * Set value of the ROW-MARKERS attribute as opposite to passed value.
    * 
    * @param    nomarkers
    *           New value of the attribute.
    */
   public void setNoRowMarkers(boolean nomarkers)
   {
      setRowMarkers(!nomarkers);
   }
   
   /**
    * Set value of the ROW-MARKERS attribute as opposite to passed value.
    * 
    * @param    nomarkers
    *           New value of the attribute.
    */
   public void setNoRowMarkers(logical nomarkers)
   {
      if (nomarkers.isUnknown())
         return;
      
      setNoRowMarkers(nomarkers.booleanValue());
   }
   
   /**
    * Set SEPARATORS attribute as opposite to passed value.
    * 
    * @param    noseparators
    *           New value for the attribute.
    */
   public void setNoSeparators(boolean noseparators)
   {
      setSeparators(!noseparators);
   }

   /**
    * Set SEPARATORS attribute as opposite to passed value.
    * 
    * @param    noseparators
    *           New value for the attribute.
    */
   public void setNoSeparators(logical noseparators)
   {
      if (noseparators.isUnknown())
         return;
      
      setNoSeparators(noseparators.getValue());
   }
   
   /**
    * Set SEPARATORS attribute.
    * 
    * @param    separators
    *           New value for the attribute.
    */
   public void setSeparators(boolean separators)
   {
      if (config.rowSeparators == separators &&
          config.columnSeparators == separators)
      {
         return;
      }

      config.rowSeparators = separators;
      config.columnSeparators = separators;
      pushWidgetAttr(new String[] {"rowSeparators", "columnSeparators"},
                     new Object[] {config.rowSeparators, config.columnSeparators});
   }

   /**
    * Set SEPARATORS attribute.
    * 
    * @param    separators
    *           New value for the attribute.
    */
   public void setSeparators(logical separators)
   {
      if (separators.isUnknown())
         return;
      
      setSeparators(separators.booleanValue());
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(boolean labels)
   {
      if (config.realized && LogicalTerminal.isChui())
      {
         errorAlreadyRealized("LABELS");
         return;
      }

      config.labels = labels;

      if (!LogicalTerminal.isChui())
      {
         for (BrowseColumnWidget column : columns)
         {
            column.config.forceLabel = false;
         }
      }

      pushScreenDefinition();
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(logical labels)
   {
      if (labels.isUnknown())
         return;
      
      setLabels(labels.booleanValue());
   }

   /**
    * Get the value of the LABELS attribute.
    *
    * @return  the value of the LABELS attribute.
    */
   public logical isLabels()
   {
      return new logical(config.labels);
   }

   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    nolabels
    *           New value for the attribute.
    */
   public void setNoLabels(boolean nolabels)
   {
      setLabels(!nolabels);
   }

   /**
    * Set CENTERED attribute value.
    * 
    * @param    centered
    *           New value for the attribute.
    */
   public void setCentered(boolean centered)
   {
      if (config.centered == centered)
      {
         return;
      }
      
      config.centered = centered;
      pushWidgetAttr("centered", config.centered);
   }

   /**
    * Set CENTERED attribute value.
    * 
    * @param    centered
    *           New value for the attribute.
    */
   public void setCentered(logical centered)
   {
      if (centered.isUnknown())
         return;
      
      setCentered(centered.booleanValue());
   }
   
   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setNoLabels(logical labels)
   {
      if (labels.isUnknown())
         return;
      
      setNoLabels(labels.booleanValue());
   }

   /**
    * Set BOX attribute.
    * 
    * @param    box
    *           New value for the attribute.
    */
   public void setBox(boolean box)
   {
      if (config.box == box)
      {
         return;
      }
      
      config.box = box;
      pushWidgetAttr("box", config.box);
   }

   /**
    * Set BOX attribute.
    * 
    * @param    box
    *           New value for the attribute.
    */
   public void setBox(logical box)
   {
      if (box.isUnknown())
         return;
      
      setBox(box.booleanValue());
   }

   /**
    * Set BOX attribute as opposite to passed value.
    * 
    * @param    nobox
    *           New value for the attribute.
    */
   public void setNoBox(boolean nobox)
   {
      setBox(!nobox);
   }

   /**
    * Set BOX attribute as opposite to passed value.
    * 
    * @param    nobox
    *           New value for the attribute.
    */
   public void setNoBox(logical nobox)
   {
      if (nobox.isUnknown())
         return;
      
      setNoBox(nobox.booleanValue());
   }

   /**
    * Set value of the MULTIPLE attribute as opposite to passed argument.
    * 
    * @param    single
    *           New value of the attribute.
    */
   public void setSingle(logical single)
   {
      if (single.isUnknown())
         return;
      
      setSingle(single.booleanValue());
   }

   /**
    * Set value of the MULTIPLE attribute as opposite to passed argument.
    * 
    * @param    single
    *           New value of the attribute.
    */
   public void setSingle(boolean single)
   {
      setMultiple(!single);
   }

   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    title
    *           The new value for the dynamic title.
    */
   public void setDynamicTitle(BaseDataType title)
   {
      setDynamicTitle(new AccessorWrapper(title));
   }
   
   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    title
    *           The new value for the dynamic title.
    */
   public void setDynamicTitle(Accessor title) 
   {
      class TitleExpr
      extends CharacterExpression
      {
         private final Accessor titleExpr;
         
         /**
          * Construct an instance for given {@link Accessor} instance.
          * 
          * @param    titleExpr
          *           Dynamic title variable.
          */
         public TitleExpr(Accessor titleExpr)
         {
            this.titleExpr = titleExpr;
         }
         
         public character execute()
         {
            return (character) titleExpr.get();
         }
      }
      setDynamicTitle(new TitleExpr(title));
   }

   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    titleExp
    *           The new expression for the dynamic title.
    */
   public void setDynamicTitle(CharacterExpression titleExp) 
   {
      this.dynTitle = titleExp;

      // the current value computed from the dynamic title is used until 
      // the buffer is sent to the client and the title is updated
      character title = dynTitle.execute();
      String    txt   = (title.isUnknown() ? "" : title.toStringMessage());
      config.title = txt;
   }

   /**
    * Set value of the NO-ASSIGN attribute.
    *
    * @param    noAssign
    *           New value of the attribute.
    */
   public void setNoAssign(boolean noAssign)
   {
      if (config.noAssign == noAssign)
      {
         return;
      }
      
      config.noAssign = noAssign;
      pushWidgetAttr("noAssign", config.noAssign);
   }

   /**
    * Set value of the NO-ASSIGN attribute.
    *
    * @param    noAssign
    *           New value of the attribute.
    */
   public void setNoAssign(logical noAssign)
   {
      if (noAssign.isUnknown())
         return;

      setNoAssign(noAssign.getValue());
   }

   /**
    * Set value of the NO-UNDERLINE attribute.
    *
    * @param    noUnderline
    *           New value of the attribute.
    */
   public void setNoUnderline(boolean noUnderline)
   {
      if (config.noUnderline == noUnderline)
      {
         return;
      }
      
      config.noUnderline = noUnderline;
      pushWidgetAttr("noUnderline", config.noUnderline);
   }

   /**
    * Set value of the NO-UNDERLINE attribute.
    *
    * @param    noUnderline
    *           New value of the attribute.
    */
   public void setNoUnderline(logical noUnderline)
   {
      if (noUnderline.isUnknown())
         return;

      setNoUnderline(noUnderline.getValue());
   }

   /**
    * Sets the NO-VALIDATE option.
    */
   public void setNoValidateOption()
   {
      config.noValidate = true;
      pushWidgetAttr("noValidate", config.noValidate);
   }

   /**
    * Sets the NO-VALIDATE writable attribute.
    *
    * @param    value
    *           The new value for the attribute.
    */
   @Override
   public void setNoValidate(boolean value)
   {
      if (!config.dynamic)
      {
         ErrorManager.recordOrShowError(4071,
               String.format("**NO-VALIDATE is an invalid attribute for a static %s %s",
                             type(), widgetName()),
               false, false);
         return;
      }
      
      if (config.noValidate == value)
      {
         return;
      }
      
      config.noValidate = value;
      pushWidgetAttr("noValidate", config.noValidate);
   }

   /**
    * Sets the NO-VALIDATE writable attribute.
    *
    * @param    value
    *           The new value for the attribute.
    */
   @Override
   public void setNoValidate(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format("**Unable to assign UNKNOWN value to attribute NO-VALIDATE on %s %s",
                             type(), widgetName()),
               false, false);
         return;
      }
      setNoValidate(value.booleanValue());
   }

   /**
    * Gets the NO-VALIDATE ttribute.
    *
    * @return   the value of the attribute.
    */
   @Override
   public logical isNoValidate()
   {
      return new logical(config.noValidate);
   }

   /**
    * Sets the AUTO-VALIDATE writable attribute.
    *
    * @param    value
    *           The new value for the attribute.
    */
   @Override
   public void setAutoValidate(boolean value)
   {
      config.autoValidate = value;
      pushWidgetAttr("autoValidate", config.autoValidate);
   }

   /**
    * Sets NO-AUTO-VALIDATE option.
    *
    * @param    noAutoValidate 
    *           The new value of the option.
    */
   public void setNoAutoValidate(boolean noAutoValidate)
   {
      setAutoValidate(!noAutoValidate);
   }

   /**
    * Sets the AUTO-VALIDATE writable attribute.
    *
    * @param    value
    *           The new value for the attribute.
    */
   @Override
   public void setAutoValidate(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute AUTO-VALIDATE on BROWSE " +
                     widgetName(),
               false);

         return;
      }

      setAutoValidate(value.getValue());
   }

   /**
    * Gets the AUTO-VALIDATE attribute.
    *
    * @return   the value of the attribute.
    */
   @Override
   public logical isAutoValidate()
   {
      return new logical(config.autoValidate);
   }

   /**
    * Get value of the FIT-LAST-COLUMN attribute.
    *
    * @return  value of the attribute.
    */
   public logical isFitLastColumn()
   {
      return new logical(config.fitLastColumn);
   }

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   public void setFitLastColumn(boolean fitLastColumn)
   {
      if (config.fitLastColumn == fitLastColumn)
      {
         return;
      }

      if (fitLastColumn && config.noEmptySpace)
      {
         ErrorManager.recordOrShowError(9956,
            String.format("To change BROWSE %s FIT-LAST-COLUMN attribute to TRUE, its " +
               "NO-EMPTY-SPACE attribute must be FALSE", name().toStringMessage()), false, false);
         return;
      }
      
      config.fitLastColumn = fitLastColumn;
      pushWidgetAttr("fitLastColumn", config.fitLastColumn);
   }

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   public void setFitLastColumn(logical fitLastColumn)
   {
      if (fitLastColumn.isUnknown())
         return;

      setFitLastColumn(fitLastColumn.getValue());
   }

   /**
    * Get value of the NO-EMPTY-SPACE attribute.
    *
    * @return  value of the attribute.
    */
   public logical isNoEmptySpace()
   {
      return new logical(config.noEmptySpace);
   }

   /**
    * Set value of the NO-EMPTY-SPACE attribute.
    *
    * @param    noEmptySpace
    *           New value of the attribute.
    */
   public void setNoEmptySpace(boolean noEmptySpace)
   {
      if (config.noEmptySpace == noEmptySpace)
      {
         return;
      }

      if (noEmptySpace && config.fitLastColumn)
      {
         ErrorManager.recordOrShowError(9962,
            String.format("To change BROWSE %s NO-EMPTY-SPACE attribute to TRUE, its " +
              "FIT-LAST-COLUMN attribute must be FALSE", name().toStringMessage()), false, false);
         return;
      }

      config.noEmptySpace = noEmptySpace;
      pushWidgetAttr("noEmptySpace", config.noEmptySpace);
   }

   /**
    * Set value of the NO-EMPTY-SPACE attribute.
    *
    * @param    noEmptySpace
    *           New value of the attribute.
    */
   public void setNoEmptySpace(logical noEmptySpace)
   {
      if (noEmptySpace.isUnknown())
         return;

      setNoEmptySpace(noEmptySpace.getValue());
   }

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   public void setExpandable(boolean expandable)
   {
      setFitLastColumn(expandable);
   }

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   public void setExpandable(logical expandable)
   {
      if (expandable.isUnknown())
         return;

      setExpandable(expandable.getValue());
   }

   /**
    * Get value of the EXPANDABLE attribute.
    *
    * @return    value of the attribute.
    */
   public logical isExpandable()
   {
      return isFitLastColumn();
   }

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnResizable(boolean resizable)
   {
      setColumnResizable(new logical(resizable));
   }

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnResizable(logical resizable)
   {
      if (resizable.isUnknown())
      {
         character name = name();
         String sName = name == null || name.isUnknown() ? "widget" : name.toStringMessage();
         String m = "**Unable to assign UNKNOWN value to attribute COLUMN-RESIZABLE on BROWSE" +
                    " %s";
         ErrorManager.recordOrShowWarning(4083, String.format(m,  sName), 
                                          false, false, false, true);   
         return;
      }

      boolean res = resizable.booleanValue();
      boolean dirty = false;
      
      // update the columns' resizable, too 
      for (BrowseColumnWidget c : columns)
      {
         if (c.config.resizable != res)
         {
            c.config.resizable = res;
            dirty = true;
         }
      }
      
      if (config.columnResizable != res)
      {
         config.columnResizable = res;
         dirty = true;
      }
      
      if (dirty)
      {
         pushScreenDefinition();
      }
   }

   /**
    * Get value of the COLUMN-RESIZABLE attribute.
    *
    * @return   <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   public logical isColumnResizable()
   {
      return new logical(config.columnResizable);
   }

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnMovable(boolean movable)
   {
      setColumnMovable(new logical(movable));
   }

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnMovable(logical movable)
   {
      if (movable.isUnknown())
         return;

      if (config.columnMovable == movable.booleanValue())
      {
         return;
      }
      
      config.columnMovable = movable.booleanValue();
      pushWidgetAttr("columnMovable", config.columnMovable);
   }

   /**
    * Get value of the CREATE-ON-ADD attribute.
    *
    * @return   value of the attribute.
    */
   public logical isCreateOnAdd()
   {
      return new logical(config.createOnAdd);
   }

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   public void setCreateOnAdd(boolean createOnAdd)
   {
      setCreateOnAdd(new logical(createOnAdd));
   }

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   public void setCreateOnAdd(logical createOnAdd)
   {
      if (createOnAdd.isUnknown())
         return;

      if (config.createOnAdd == createOnAdd.booleanValue())
      {
         return;
      }
      
      config.createOnAdd = createOnAdd.booleanValue();
      pushWidgetAttr("createOnAdd", config.createOnAdd);
   }

   /**
    * Get value of the COLUMN-MOVABLE attribute.
    *
    * @return   <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   public logical isColumnMovable()
   {
      return new logical(config.columnMovable);
   }

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   public void setRefreshable(boolean refreshable)
   {
      setRefreshable(new logical(refreshable));
   }

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   public void setRefreshable(logical refreshable)
   {
      if (refreshable.isUnknown())
         return;

      boolean val = refreshable.booleanValue();
      if (config.refreshable == val)
         return;

      config.refreshable = val;
      pushWidgetAttr("refreshable", config.refreshable);

      if (val)
      {
         boolean queryReopened = pendingReopen == PendingOpenState.REOPEN ||
                                 pendingReopen == PendingOpenState.REPOSITIONED_AND_REOPEN;
         if (queryReopened)
            maximumKnownRowIndex = null;

         LogicalTerminal.executePendingRefresh(this,
                                       queryReopened,
                                       pendingReopen == PendingOpenState.REPOSITIONED_AND_REOPEN,
                                       pendingRepositionRow);
      }
      else
      {
         pendingReopen = PendingOpenState.NO_REOPEN;
         pendingRepositionRow = -1;
      }
   }

   /**
    * Get value of the REFRESHABLE attribute.
    *
    * @return   <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   public logical isRefreshable()
   {
      return new logical(config.refreshable);
   }

   /**
    * Set value of the ROW-RESIZABLE attribute.
    *
    * @param    flag
    *           <code>true</code> if the row height can be changed.
    */
   @Override
   public void setRowResizable(boolean flag)
   {
      config.rowResizable = flag;
      pushWidgetAttr("rowResizable", config.rowResizable);
   }

   /**
    * Get value of the ROW-RESIZABLE attribute.
    *
    * @return   <code>true</code> if the row height can be changed.
    */
   @Override
   public logical isRowResizable()
   {
      return new logical(config.rowResizable);
   }

   /**
    * Get value of the CURRENT-ROW-MODIFIED attribute.
    *
    * @return   <code>true</code> if any cells in the current row have been changed.
    */
   public logical isCurrentRowModified()
   {
      return new logical(LogicalTerminal.isCurrentRowModified(this));
   }

   /**
    * Get value of the NEW-ROW attribute.
    *
    * @return   <code>true</code> if the row in focus was added to the browse using the
    *           INSERT-ROW() method and has not been added to the database.
    */
   public logical isNewRow()
   {
      return new logical(LogicalTerminal.isNewRow(this));
   }

   /**
    * Get value of the ROW-MARKERS attribute.
    *
    * @return   <code>true</code> if the browse has row markers.
    */
   public logical hasRowMarkers()
   {
      if (config.rowMarkers != null)
      {
         return new logical(config.rowMarkers);
      }
      else if (config.dynamic)
      {
         return new logical(true);
      }

      boolean editable = false;
      for (BrowseColumnWidget column : columns)
      {
         if (!column.isReadOnly().booleanValue())
         {
            editable = true;
            break;
         }
      }
      return new logical(editable);
   }

   /**
    * Gets the LABEL-BGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-BGCOLOR attribute.
    */
   public integer getLabelBgColor()
   {
      int color = config.labelBgColor;
      return (color == -1) ? new integer() : new integer(color);
   }

   /**
    * Gets the LABEL-FGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-FGCOLOR attribute.
    */
   public integer getLabelFgColor()
   {
      int color = config.labelBgColor;
      return (color == -1) ? new integer() : new integer(color);
   }

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   @Override
   public void setLabelBgColor(int64 bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setLabelBgColor(color);
   }
   
   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   @Override
   public void setLabelBgColor(long bgcolor)
   {
      if (config.labelBgColor == bgcolor)
      {
         return;
      }
      
      config.labelBgColor = (int) bgcolor;
      pushWidgetAttr("labelBgColor", config.labelBgColor);
   }

   /**
    * Sets the LABEL-FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the LABEL-FGCOLOR attribute.
    */
   public void setLabelFgColor(int64 fgcolor)
   {
      int color = (fgcolor == null || fgcolor.isUnknown()) ? -1 : fgcolor.intValue();
      setLabelFgColor(color);
   }

   /**
    * Sets the LABEL-FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the LABEL-FGCOLOR attribute.
    */
   public void setLabelFgColor(long fgcolor)
   {
      if (config.labelFgColor == fgcolor)
      {
         return;
      }

      config.labelFgColor = (int) fgcolor;
      pushWidgetAttr("labelFgColor", config.labelFgColor);
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute.
    *
    * @param    columnScrolling
    *           New value of the attribute.
    */
   public void setColumnScrolling(boolean columnScrolling)
   {
      if (config.columnScrolling == columnScrolling)
      {
         return;
      }
      
      config.columnScrolling = columnScrolling;
      pushWidgetAttr("columnScrolling", config.columnScrolling);
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute.
    *
    * @param    columnScrolling
    *           New value of the attribute.
    */
   public void setColumnScrolling(logical columnScrolling)
   {
      if (columnScrolling.isUnknown())
         return;

      setColumnScrolling(columnScrolling.getValue());
   }

   /**
    * Get value of the COLUMN-SCROLLING attribute.
    *
    * @return   <code>true</code> if browse is scrolled horizontally per column.
    *           <code>false</code> for usual scrolling.
    */
   public logical isColumnScrolling()
   {
      return new logical(config.columnScrolling);
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute as opposite to passed argument.
    *
    * @param    noColumnScrolling
    *           New value of the attribute.
    */
   public void setNoColumnScrolling(boolean noColumnScrolling)
   {
      if (config.columnScrolling == !noColumnScrolling)
      {
         return;
      }
      
      config.columnScrolling = !noColumnScrolling;
      pushWidgetAttr("columnScrolling", config.columnScrolling);
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute as opposite to passed argument.
    *
    * @param    noColumnScrolling
    *           New value of the attribute.
    */
   public void setNoColumnScrolling(logical noColumnScrolling)
   {
      if (noColumnScrolling.isUnknown())
         return;

      setNoColumnScrolling(noColumnScrolling.getValue());
   }

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   public void setRowHeightChars(double rowHeightChars)
   {
      setRowHeightChars(new decimal(rowHeightChars));
   }

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   public void setRowHeightChars(NumberType rowHeightChars)
   {
      Double val = rowHeightChars.isUnknown() ? null : rowHeightChars.doubleValue();
      if (val == null)
      {
         rowHeightError(4083, false);
         return;
      }
      else if (val < 0)
      {
         rowHeightError(4060, false);
         return;
      }
      else if (val > 0 && _isRealized())
      {
         int pixelVal = LogicalTerminal.coordinates().rowToPixels(val);
         if (LogicalTerminal.calcBrowseDown(this, pixelVal) < MIN_BROWSE_ROWS)
         {
            rowHeightError(6422, false);
            rowHeightError(3131, false);
            return;
         }
      }

      CharsPixels charsPixels = val == 0 ?
             new CharsPixels(0, 0) : // ROW-HEIGHT-CHARS = 0 is a valid value
                                     // (but not ROW-HEIGHT-PIXELS = 0)
             BaseEntity.prepareSizeChars(val, SizeAttribute.ROW_HEIGHT_CHARS, this);
      if (charsPixels == null)
      {
         return;
      }

      config.rowHeightChars  = charsPixels.chars;
      config.rowHeightPixels = charsPixels.pixels;
      pushWidgetAttr(new String[] {"rowHeightChars", "rowHeightPixels"},
                     new Object[] {config.rowHeightChars, config.rowHeightPixels});
   }

   /**
    * Set value of the ROW-HEIGHT-PIXELS attribute.
    *
    * @param    rowHeightPixels
    *           New value of the attribute.
    */
   public void setRowHeightPixels(int rowHeightPixels)
   {
      setRowHeightPixels(new integer(rowHeightPixels));
   }

   /**
    * Set value of the ROW-HEIGHT-PIXELS attribute.
    *
    * @param    rowHeightPixels
    *           New value of the attribute.
    */
   public void setRowHeightPixels(NumberType rowHeightPixels)
   {
      setRowHeightPixels(rowHeightPixels, true);
   }

   /**
    * Set value of the ROW-HEIGHT-PIXELS attribute.
    *
    * @param    rowHeightPixels
    *           New value of the attribute.
    * @param    push
    *           If ROW-HEIGHT-* attributes are immediately pushed to the client side.
    */
   public void setRowHeightPixels(NumberType rowHeightPixels, boolean push)
   {
      Integer val = rowHeightPixels.isUnknown() ? null : rowHeightPixels.intValue();
      if (val == null)
      {
         rowHeightError(4083, true);
         return;
      }
      else if (_isRealized() &&
               (val < 0 || LogicalTerminal.calcBrowseDown(this, val) < MIN_BROWSE_ROWS))
      {
         rowHeightError(6422, true);
         rowHeightError(3131, true);
         return;
      }

      // rowHeightPixels = 0 case is handled in prepareSizePixels
      CharsPixels charsPixels = !_isRealized() && val < 0 ?
                       new CharsPixels(-1, -1) : //TODO actually you can assign negative
                                                 // ROW-HEIGHT-PIXELS before widget is realized,
                                                 // so far we reset values to default
                       BaseEntity.prepareSizePixels(val, SizeAttribute.ROW_HEIGHT_PIXELS, this);
      if (charsPixels == null)
      {
         return;
      }

      config.rowHeightChars  = charsPixels.chars;
      config.rowHeightPixels = charsPixels.pixels;

      if (push)
      {
         pushWidgetAttr(new String[]{"rowHeightChars", "rowHeightPixels"},
                        new Object[]{config.rowHeightChars, config.rowHeightPixels});
      }
   }

   /**
    * Implements the ROW-HEIGHT-PIXELS attribute getter.
    *
    * @return row height, in pixels, of a browse row.
    */
   public integer getRowHeightPixels()
   {
      return new integer(config.rowHeightPixels);
   }

   /**
    * Sets the TITLE writable attribute.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   public void setTitle(String title)
   {
      if (config.realized && "".equals(config.title))
      {
         // We are free to change TITLE of a realized browse with a title. But we cannot set a
         // TITLE for a realized browse without title.
         errorAlreadyRealized("TITLE");
         return;
      }

      super.setTitle(title);
      
      // if the browse TITLE attribute is set, then the dynamic title
      // must be canceled
      this.dynTitle = null;
   }

   /**
    * Get value of NUM-COLUMNS attribute.
    * 
    * @return   Number of columns in the browse.
    */
   public integer getNumColumns()
   {
      return new integer(columns.size());
   }

   /**
    * Get specified browse column.
    *
    * @param    index
    *           1-based column index.
    *
    * @return   reference to column widget.
    */
   public handle getBrowseColumn(int index)
   {
      return getBrowseColumn(index, true);
   }

   /**
    * Get specified browse column.
    * 
    * @param    index
    *           1-based column index.
    * @param    showError
    *           indicates if error should be reported.
    *           
    * @return   reference to column widget.
    */
   public handle getBrowseColumn(int index, boolean showError)
   {
      BrowseColumnWidget res = null;
               
      if (index >= 1 && index <= columns.size())
      {
         res = columns.get(index - 1);
      }

      if (res == null)
      {
         if (showError)
            ErrorManager.recordOrShowError(4065,
                "**The GET-BROWSE-COLUMN attribute on the BROWSE " +
                widgetName() + " has invalid arguments.",
             true);
         return new handle();
      }
      
      return res.asWidgetHandle();
   }

   /**
    * Get specified browse column.
    * 
    * @param    index
    *           1-based column index.
    *           
    * @return   reference to column widget.
    */
   public handle getBrowseColumn(NumberType index)
   {
      int ndx = (index.isUnknown()) ? 0 : index.intValue();
      
      return getBrowseColumn(ndx);
   }
   
   /**
    * Get the number of rows currently visible in a browse widget.
    *
    * @return  the number of rows currently visible in a browse widget.
    */
   @Override
   public integer visibleIterations()
   {
      return config.realized ? getDown() : new integer();
   }
   
   /**
    * Get the NUM-SELECTED-ROWS attribute.
    * 
    * @return   The current value of NUM-SELECTED-ROWS attribute.
    */
   public integer getNumSelectedRows()
   {
      return new integer(LogicalTerminal.getNumSelectedRows(this));
   }

   /**
    * Set dynamic filtering for the browse.
    *
    * @param   fieldRefs
    *          Array of the references to the fields to filter by.
    * @param   values
    *          Array of the filter values.
    * @param   formats
    *          Array of formats used to display the values.
    */
   public void setDynamicFiltering(FieldReference[] fieldRefs,
                                   BaseDataType[] values,
                                   String[] formats)
   {
      if (query == null)
      {
         return;
      }
      boolean wasOpen = query.isOpen().booleanValue();
      query.close();
      
      query.clearDynamicFilters();
      if (fieldRefs != null && fieldRefs.length > 0)
      {
         for (int i = 0; i < fieldRefs.length; i++)
         {
            query.addDynamicFilter(fieldRefs[i], values[i], formats[i]);
         }
      }
      
      if (wasOpen)
      {
         query.open();
      }
   }

   /**
    * Set dynamic sorting for this browse.
    *
    * @param fieldRefs
    *        Array of the references to the fields to sort by. Inside the array reference are
    *        positioned from the reference which defines the most coarse sorting to the most
    *        granular.
    * @param ascending
    *        Array of the sorting directions. <code>true</code> for ascending, <code>false</code>
    *        for descending. Inside the array directions are positioned from the field reference
    *        which defines the most coarse sorting to the most granular.
    */
   public void setDynamicSorting(FieldReference[] fieldRefs, boolean[] ascending)
   {
      if (query == null)
      {
         return;
      }
      boolean wasOpen = query.isOpen().booleanValue();
      query.close();

      query.clearDynamicSorts();
      if (fieldRefs != null && fieldRefs.length > 0)
      {
         if (originalQuery == null)
         {
            originalQuery = query;

            QueryWrapper newQuery = query.createDynamicSorter(fieldRefs, ascending);

            if (wasOpen)
            {
               newQuery.open();
            }

            setQuery(newQuery);
         }
         else
         {
            query.addDynamicSort(fieldRefs, ascending);

            if (wasOpen)
            {
               query.open();
            }
         }
      }
      else if (wasOpen)
      {
         query.open();
      }
   }
   
   /**
    * Get the QUERY writable attribute.
    * 
    * @return   Current value of the QUERY attribute.
    */
   @Override
   public P2JQuery getQuery()
   {
      return query;
   }
   
   /**
    * Set the query which will back this browse.  Ensure it is scrollable and
    * does not raise an error condition if it goes off-end.
    * 
    * @param    query
    *           New value for the attribute. <code>null</code> to unbind previous query.
    */
   @Override
   public void setQuery(P2JQuery query)
   {
      setQuery(query, false);
   }
   
   /**
    * Callback for the query to do any finilazing actions in browse widget when the query opened.
    */
   @Override
   public void afterQueryOpen()
   {
      // the query must be valid
      if (query == null || !deferQueryReposition)
      {
         return;
      }
      
      try
      {
         // defer reposition for some cases
         query.reposition(1);
      }
      catch (QueryOffEndException ex)
      {
         // result set is empty.
      }
   }

   /**
    * Set the query which will back this browse.  Ensure it is scrollable and
    * does not raise an error condition if it goes off-end.
    * 
    * @param    query
    *           New value for the attribute. <code>null</code> to unbind previous query.
    * @param    deferReposition
    *           When <code>TRUE</code> postpone query reposition until openong, when
    *           <code>FALSE</code> reposition immediately.
    */
   private void setQuery(P2JQuery query, boolean deferReposition)
   {
      if (query != null)
      {
         if (query.equals(this.query))
         {
            // same query, do nothing
            return;
         }
         
         boolean error = false;
         if (!query.isScrolling())
         {
            ErrorManager.recordOrShowError(3330,
                  "BROWSE requires the QUERY to be DEFINEd SCROLLING", false, false);
            error = true;
         }
         
         if (query.isBrowsed())
         {
            ErrorManager.recordOrShowError(3147,
                  String.format("Cannot browse query %s because query is already being browsed",
                                ((QueryWrapper) query).getName()),
                  false, false);
            error = true;
         }

         if (error)
         {
            ErrorManager.recordOrShowError(4078,
                  "**Unable to set attribute QUERY for BROWSE widget", false, false);
            return;
         }
      }

      // The rule to clear the browse columns when query is assigned seems to be: (1) for dynamic browse
      // assigning ? or any query not already assigned (not having the same resource id), (2) for
      // non-dynamic browse when ? is assigned or a query with incompatible buffers assigned.
      // Incompatible buffers for two queries are those that don't match the data types of the underlying
      // buffer fields (i.e. different count, or different data types of the fields at the same sequence
      // number). This holds even for buffers of different tables, that is buffers of different tables can
      // match as long as the fields match.
      if (config.dynamic || (this.query != null && !matchBufferFields(this.query, (QueryWrapper) query)))
      {
         columns = new ArrayList<>();
         updateColumnIds();
         pushWidgetAttr("columnIds", config.columnIds);
         notifyColumnsUpdated(true);
      }

      // Unbind previous query.
      if (this.query != null)
      {
         this.query.setBrowsed(false);
         this.query.setFetchOnReposition(false);
         this.query.removeRepositionListener(this);
         // remove listener for browse widget
         this.query.removeOpenedListener(this);

         this.query = null;
         this.deferQueryReposition = false;
      }
      
      if (query == null)
      {
         // no new query to set
         queryClosed();
         return;
      }

      this.query = (QueryWrapper) query;
      this.query.setBrowsed(this);
      query.setFetchOnReposition(true);
      query.addRepositionListener(this);
      
      try
      {
         // defer reposition for some cases
         if (deferReposition)
         {
            this.query.addOpenedListener(this);
         }
         else
         {
            query.reposition(1);
         }
      }
      catch (QueryOffEndException ex)
      {
         // result set is empty.
      }
      
      initMaxDataGuess();

      config.editPossible = calcEditPossible();
      config.queryAssigned = true;

      // Create widgets for editing.
      createEditors();
      pushScreenDefinition();
   }

   /**
    * Get the query associated with the given resource.
    *
    * @return   The associated query or <code>unknown</code> if none exists.
    */
   public handle getQueryAsHandle()
   {
      P2JQuery query = getQuery();
      
      return (query == null) ? new handle() : new handle((QueryWrapper) query);
   }
   
   /**
    * Set the query associated with the given resource.
    *
    * @param    qry
    *           The query to associate.
    */
   @Override
   public void setQueryAsHandle(handle qry)
   {
      if (qry == null || qry.isUnknown() || !qry._isValid())
      {
         setQuery(null);
         return;
      }
      
      Object resource = qry.get();
      
      if (!(resource instanceof QueryWrapper))
      {
         ErrorManager.recordOrShowError(4062,
               String.format(
                     "**Attribute QUERY for the BROWSE widget was passed an invalid %s widget",
                     ((HandleResource) resource).type()),
               false, false);
         return;
      }
      
      // find out if we need deferred reposition for the query was not opened
      deferQueryReposition = this.query == null && _dynamic() && resource != null &&
                             !((QueryWrapper)resource).isOpen().booleanValue();
                                
      setQuery((QueryWrapper) resource, deferQueryReposition);
   }

   /**
    * Set the query associated with the given resource.
    *
    * @param    qry
    *           The query to associate.
    */
   @Override
   public void setQueryAsHandle(QueryWrapper qry)
   {
      setQuery(qry);
   }

   /**
    * Implements the SELECT-ALL() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical selectAll()
   {
      if (!isMultiple().booleanValue())
      {
         final String msg = "You cannot set SELECT-ALL for a single select BROWSE brws";
         ErrorManager.recordOrShowError(4674, msg, false, false);
         return new logical(false);
      }

      LogicalTerminal.selectAll(this);
      return new logical(true);
   }

   /**
    * Implements the DESELECT-ROWS() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical deselectRows()
   {
      if (!checkNoRecords("DESELECT-ROWS"))
      {
         return new logical(false);
      }

      if (frame == null) 
      {
         String[] msg = new String[]
         { "Unable to realise BROWSE widget because it is not in a frame",
           "Unknown error code 1 for attribute ? on the BROWSE widget",
            "**Unable to query attribute DESELECT-ROWS for BROWSE widget" };
         int[] num = new int[] {4040, 4104, 4077};
         ErrorManager.recordOrShowError(num, msg, false);
         return new logical(false);
      }
      
      LogicalTerminal.deselectRows(this);
      return new logical(true);
   }

   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   public logical fetchSelectedRow(long n)
   {
      return fetchSelectedRow(new int64(n));
   }
   
   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   public logical fetchSelectedRow(int64 n)
   {
      if (query == null)
      {
         final String msg = "Browse widget must contain at least one record" + 
                            " before FETCH-SELECTED-ROW method is used";
         ErrorManager.displayError(2108, msg, false);

         return new logical(false);
      }

      if (n == null || n.isUnknown())
      {
         final String msg = "Invalid row index specified for browse method " + 
                            "FETCH-SELECTED-ROW";
         ErrorManager.displayError(385, msg, false);

         return new logical(false);
      }
      
      int selListIndex = n.intValue() - 1;
      if (selListIndex < 0)
         return new logical(false);
      
      int uiRow;
      try
      {
         // Get the 0-based row index of the (n - 1)th selected row.
         uiRow = LogicalTerminal.getSelectedRowIndex(this, selListIndex);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, "FETCH-SELECTED-ROW");
         return new logical(false);
      }

      if (uiRow < 0)
      {
         return new logical(false);
      }
      
      // Reposition the query to this row and update the underlying buffers.
      reposition(new integer(uiRow + 1), true);
      
      if (query.isOffEnd().booleanValue())
         return new logical(false);
      
      return new logical(true);
   }

   /**
    * Implements the REFRESH() widget method.
    *
    * @return  Always <code>true</code>.
    */
   public logical refresh()
   {
      return refresh(false);
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    *
    * @param   deleteCurrentRowMode
    *          <code>true</code> if this function is used for update on DELETE-CURRENT-ROW.
    *          <code>false</code> if this function is used for update on REFRESH.
    *
    * @return  Always <code>true</code>.
    */
   public logical refresh(boolean deleteCurrentRowMode)
   {
      LogicalTerminal.refresh(this, deleteCurrentRowMode);
      return new logical(true);
   }

   /**
    * Implements the SELECT-FOCUSED-ROW() widget method.<p>
    * Selects the row that currently has focus in a browse widget, even if it is not currently
    * displayed.
    * 
    * @return   Always <code>true</code>.
    */
   public logical selectFocusedRow()
   {
      LogicalTerminal.selectFocusedRow(this);
      return new logical(true);
   }

   /**
    * Implements the DESELECT-FOCUSED-ROW() widget method.<p>
    * Deselects the row with current focus.
    *
    * @return   <code>true</code> if the browse has at least one record.
    */
   public logical deselectFocusedRow()
   {
      if (!checkNoRecords("DESELECT-FOCUSED-ROW"))
      {
         return new logical(false);
      }

      LogicalTerminal.deselectFocusedRow(this);
      return new logical(true);
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   @Override
   public logical selectRow(int64 row)
   {
      if (!checkNoRecords("SELECT-ROW"))
      {
         return new logical(false);
      }

      Integer val = row.isUnknown() ? null : row.intValue();
      if (val == null ||
          val < 1     ||
          val > getBrowsePageRowCount() ||
          !LogicalTerminal.selectRow(this, val))
      {
         errorInvalidRowIndex("SELECT-ROW");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   @Override
   public logical selectRow(long row)
   {
      return selectRow(new int64(row));
   }

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public logical selectNextRow()
   {
      final String methodName = "SELECT-NEXT-ROW";
      if (!checkNoRecords(methodName))
      {
         return new logical(false);
      }

      try
      {
         return LogicalTerminal.selectNextRow(this);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
         return new logical(); //correct
      }
   }

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public logical selectPreviousRow()
   {
      final String methodName = "SELECT-PREV-ROW";
      if (!checkNoRecords(methodName))
      {
         return new logical(false);
      }

      try
      {
         return LogicalTerminal.selectPreviousRow(this);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
         return new logical(); //correct
      }
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   public logical deselectSelectedRow(int64 row)
   {
      if (!checkNoRecords("DESELECT-SELECTED-ROW"))
      {
         return new logical(false);
      }

      Integer val = row.isUnknown() ? null : row.intValue();
      if (val == null || val < 1 || !LogicalTerminal.deselectSelectedRow(this, val))
      {
         errorInvalidRowIndex("DESELECT-SELECTED-ROW");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   public logical deselectSelectedRow(int row)
   {
      return deselectSelectedRow(new integer(row));
   }

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   public logical deleteSelectedRow(int64 row)
   {
      final String methodName = "DELETE-SELECTED-ROW";
      if (row.isUnknown())
      {
         errorInvalidRowIndex(methodName);
         return new logical(false);
      }

      return deleteSingleRow(row.intValue() - 1, methodName);
   }

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   public logical deleteSelectedRow(int row)
   {
      return deleteSelectedRow(new integer(row));
   }

   /**
    * Implements the DELETE-SELECTED-ROWS() widget method.<p>
    * Deletes all currently selected rows from a browse and the associated results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if rows were successfully deleted.
    */
   public logical deleteSelectedRows()
   {
      final String methodName = "DELETE-SELECTED-ROWS";

      int[] selectedRows;
      try
      {
         selectedRows = LogicalTerminal.getSelectedRowIndexes(this);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
         return new logical(false);
      }

      Arrays.sort(selectedRows);

      for (int i = selectedRows.length - 1; i >= 0; i--)
      {
         logical res = deleteRow(selectedRows[i], false);
         if (!res.booleanValue())
            return res;
      }

      LogicalTerminal.refreshAfterDeletion(this, -1);
      return new logical(true);
   }

   /**
    * Implements the DELETE-CURRENT-ROW() widget method.<p>
    * Deletes the most recently selected row from a browse and the results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   public logical deleteCurrentRow()
   {
      return deleteSingleRow(-1, "DELETE-CURRENT-ROW");
   }

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   public logical isRowSelected(int64 row)
   {
      if (!checkNoRecords("IS-ROW-SELECTED"))
      {
         return new logical(false);
      }

      Integer val = row.isUnknown() ? null : row.intValue();
      if (val == null                   ||
          val < 1                       ||
          val > getBrowsePageRowCount() ||
          val > query.getNumResults().intValue())
      {
         errorInvalidRowIndex("IS-ROW-SELECTED");
         return new logical(false);
      }

      boolean res = LogicalTerminal.isRowSelected(this, val);
      return new logical(res);
   }

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   public logical isRowSelected(int row)
   {
      return isRowSelected(new integer(row));
   }


   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   public logical insertRow(character mode)
   {
      if (!isEditPossible())
      {
         ErrorManager.recordOrShowError(4527,
                                        "Cannot use INSERT-ROW in READ-ONLY browser",
                                        false,
                                        false);
         return new logical(false);
      }

      boolean after;
      if (mode == null || mode.isUnknown())
         after = false;
      else
      {
         String normalizedValue = TextOps.rightTrim(mode).getValue().toUpperCase();
         switch (normalizedValue)
         {
            case "AFTER":
               after = true;
               break;
            case "BEFORE":
               after = false;
               break;
            default:
               ErrorManager.recordOrShowError(4065,
                  String.format("The INSERT-ROW attribute on the BROWSE %s has invalid arguments",
                                name().getValue()),
                     false, true);
               return new logical(false);
         }
      }

      boolean res = false;
      try
      {
         res = LogicalTerminal.insertRow(this, after);
      }
      catch (ErrorConditionException e)
      {
         if (e.getProgressErrorCode() == 4963)
         {
            ErrorManager.recordOrShowError(e.getProgressErrorCode(),
             String.format("You must have a database row in BROWSE %s in order to use INSERT-ROW",
                           name().getValue()),
                  false, false);
         }
         else
         {
            throw new IllegalStateException("Unknown error code: " + e.getProgressErrorCode());
         }
      }

      return new logical(res);
   }

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   public logical insertRow(String mode)
   {
      return insertRow(new character(mode));
   }

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   public logical insertRow()
   {
      return insertRow(new character("BEFORE"));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(String fieldName, int64 position)
   {
      return addLikeColumn(new character(fieldName), position);
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(character fieldName, int64 position)
   {
      if (!checkQueryAssigned())
      {
         return new handle();
      }

      FieldReference[] fieldReferences = DBUtils.getFieldForAddLikeColumn(query, fieldName);
      return addLikeColumn(fieldReferences, position);
   }

   /**
    * Implements ADD-LIKE-COLUMN() widget method. Creates browse column(s) from the specified
    * field(s).
    *
    * @param  fieldReferences
    *         Reference(s) to the target field(s).
    * @param  position
    *         1-based position of the new browse columns.
    *
    * @return handle to the first created column.
    */
   public handle addLikeColumn(FieldReference[] fieldReferences, int64 position)
   {
      //TODO there is quirk in ChUI version that an empty result set can be displayed after this
      // function. e.g. if quality operator is used in the query predicate.

      BrowseColumnWidget resultColumnWidget = null;
      if (fieldReferences != null)
      {
         for (FieldReference fieldReference : fieldReferences)
         {
            Buffer buffer = (Buffer) fieldReference.getDMO();
            String property = fieldReference.getProperty();

            String columnLabel;
            int extentIndex = fieldReference.getIndex();

            // TODO currently P2J returns an empty string if a column label is not specified,
            // so it is not clear how distinguish cases when column label is not specified and
            // when an empty one is specified
            columnLabel = TableMapper.getLegacyFieldColumnLabel(buffer, property);
            if (columnLabel == null || columnLabel.isEmpty()) // TODO fix
               columnLabel = TableMapper.getLegacyFieldLabel(buffer, property);
            if (columnLabel == null || columnLabel.isEmpty()) // TODO fix
               columnLabel = TableMapper.getLegacyFieldName(buffer, property);

            if (extentIndex != -1)
               columnLabel = String.format("%s[%d]", columnLabel, extentIndex + 1);

            String dataType   = fieldReference.getType().getSimpleName();
            String dataFormat = TableMapper.getLegacyFieldFormat(buffer, property);

            // get VALIDATE-EXPRESSION through BufferField
            String legacyName = TableMapper.getLegacyFieldName(buffer, property);
            BufferField bufferField = buffer.bufferField(legacyName).unwrapBufferField();
            String valExpr = bufferField.validateExpression().getValue();
            String valMsg  = bufferField.validateMessage().getValue();

            BrowseColumnWidget columnWidget = addColumnWidget(dataType,
                                                              dataFormat,
                                                              columnLabel,
                                                              position,
                                                              fieldReference,
                                                              null,
                                                              FILL_IN,
                                                              valExpr,
                                                              valMsg,
                                                              false);

            if (resultColumnWidget == null)
            {
               resultColumnWidget = columnWidget;
            }
         }
      }
      return resultColumnWidget == null ? new handle() : resultColumnWidget.asWidgetHandle();
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(String fieldName, long position)
   {
      return addLikeColumn(new character(fieldName), new integer(position));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse's query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(handle fieldHandle, int64 position)
   {
      if (!checkQueryAssigned())
      {
         return new handle();
      }

      FieldReference[] fieldReferences = DBUtils.getFieldForAddLikeColumn(query, fieldHandle);
      return addLikeColumn(fieldReferences, position);
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse's query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(handle fieldHandle, long position)
   {
      return addLikeColumn(fieldHandle, new integer(position));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(character fieldName)
   {
      return addLikeColumn(fieldName, new integer(0));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(String fieldName)
   {
      return addLikeColumn(new character(fieldName));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse's query.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(handle fieldHandle)
   {
      return addLikeColumn(fieldHandle, new integer(0));
   }

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   public logical moveColumn(int64 src, int64 dest)
   {
      if (LogicalTerminal.isChui())
      {
         UnimplementedFeature.partial("MOVE-COLUMN is not implemented for ChUI");
         return new logical(false);
      }

      boolean valid = !src.isUnknown() && !dest.isUnknown();

      int srcInt = 0;
      int destInt = 0;

      if (valid)
      {
         srcInt = src.intValue() - 1; // 0-based
         destInt = dest.intValue() - 1; // 0-based

         if (srcInt == destInt)
         {
            return new logical(true);
         }

         int colNum = columns.size();

         valid = srcInt >= 0 && srcInt < colNum &&
                 destInt >= 0 && destInt < colNum;
      }

      if (!valid)
      {
         ErrorManager.recordOrShowError(4077,
               "**Unable to query attribute MOVE-COLUMN for BROWSE " +
                     name().toStringMessage(),
               true);
         return new logical(false);
      }

      if (destInt > srcInt)
      {
         Collections.rotate(columns.subList(srcInt, destInt + 1), -1);
      }
      else
      {
         Collections.rotate(columns.subList(destInt, srcInt + 1), 1);
      }
      updateColumnIds();
      pushScreenDefinition();
      // ordinals are re-calculated on the client side to avoid flickering on pushScreenDefinition
      LogicalTerminal.moveColumnClient(this, srcInt, destInt);
      return new logical(true);
   }

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved (including invisible columns).
    * @param  dest
    *         1-based position to which the column is moved (including invisible columns).
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   public logical moveColumn(long src, long dest)
   {
      return moveColumn(new int64(src), new int64(dest));
   }
   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   public logical moveColumn(long src, int64 dest)
   {
      return moveColumn(new int64(src), dest);
   }
   
   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   public logical moveColumn(int64 src, long dest)
   {
      return moveColumn(src, new int64(dest));
   }

   /**
    * Implements the GET-REPOSITIONED-ROW() widget method.<p>
    * Returns the row index of the browse viewport where the REPOSITION TO ROWID (or
    * RECID) statement displays a repositioned record.
    *
    * @return the row index of the browse viewport where the REPOSITION TO ROWID (or
    *         RECID) statement displays a repositioned record
    */
   public integer getRepositionedRow()
   {
      return new integer(config.repositionedRow + 1);
   }

   /**
    * Implements the FIRST-COLUMN attribute getter.<p>
    *
    * @return a handle to the first column in a browse widget, regardless of the value of its
    *         READ-ONLY attribute or its VISIBLE attribute.
    */
   public handle getFirstColumn()
   {
      if (getNumColumns().intValue() == 0)
         return new handle();

      return getBrowseColumn(1);
   }

   /**
    * Implements the FOCUSED-ROW attribute getter.<p>
    *
    * @return the 1-based index of the focused row in the viewport.
    */
   public integer getFocusedRow()
   {
      return new integer(LogicalTerminal.getFocusedRow(this));
   }

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute getter.<p>
    *
    * @return the number of visible leading columns locked in a browse widget.
    */
   public integer getNumLockedColumns()
   {
      return new integer(config.numLockedColumns);
   }

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   public void setNumLockedColumns(double num)
   {
      setNumLockedColumns(new decimal(num));
   }

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   public void setNumLockedColumns(NumberType num)
   {
      if (num.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute NUM-LOCKED-COLUMNS on BROWSE " +
                     name().toStringMessage(),
               false);
         return;
      }

      int val = num.intValue();

      if (val < 0)
      {
         ErrorManager.displayError(4082,
               "**Attribute NUM-LOCKED-COLUMNS must be greater than or equal to 0 on BROWSE " +
                     name().toStringMessage(),
               false);
         return;
      }

      if (val > columns.size())
      {
         ErrorManager.displayError(4533,
               "You cannot set NUM-LOCKED-COLUMNS to be greater than the total number of " +
                  "columns for BROWSE " + name().toStringMessage(),
               false);
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute NUM-LOCKED-COLUMNS in widget " +
                  name().toStringMessage() + " of type BROWSE",
               false);
         return;
      }

      if (LogicalTerminal.isChui() && initDone)
      {
         // In ChUI you cannot change the value of the attribute after browse was displayed.
         ErrorManager.displayError(4088,
               "**Progress does not support the NUM-LOCKED-COLUMNS attribute for the BROWSE " +
                  name().toStringMessage() + " in this display environment",
               false);
         return;
      }

      config.numLockedColumns = val;
      pushWidgetAttr("numLockedColumns", config.numLockedColumns);
   }

   /**
    * Implements the NUM-VISIBLE-COLUMNS attribute getter.<p>
    *
    * @return the number of visible columns browse widget.
    */
   @Override
   public integer getNumVisibleColumns()
   {
      int retNumVis = 0;
      
      for (BrowseColumnWidget column : columns)
      {
         if (column._isVisible())
         {
            retNumVis++;
         }
      }
      
      return new integer(retNumVis);
   }

   /**
    * Implements the MAX-HEIGHT-CHARS attribute getter.<p>
    *
    * @return  maximum height of the browse, in character units.
    */
   @Override
   public decimal getMaxHeightChars()
   {
      ErrorManager.displayError(4052, "MAX-HEIGHT is not a queryable attribute for BROWSE "+
                                widgetName());
      return new decimal();
   }

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   @Override
   public void setMaxHeightChars(NumberType max)
   {
      setMaxHeightChars(new decimal(max));
   }

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   @Override
   public void setMaxHeightChars(decimal max)
   {
      ErrorManager.displayError(4052, "MAX-HEIGHT is not a setable attribute for BROWSE "+
                                widgetName());
   }

   /**
    * Implements the MIN-HEIGHT-CHARS attribute getter.<p>
    *
    * @return  minimum height of the browse, in character units.
    */
   @Override
   public decimal getMinHeightChars()
   {
      return new decimal(config.minHeightChars);
   }

   /**
    * Sets the MIN-HEIGHT-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-HEIGHT-CHARS attribute.
    */
   @Override
   public void setMinHeightChars(NumberType min)
   {
      setMinHeightChars(new decimal(min));
   }

   /**
    * Sets the MIN-HEIGHT-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-HEIGHT-CHARS attribute.
    */
   @Override
   public void setMinHeightChars(decimal min)
   {
      ErrorManager.displayError(4052, "MIN-HEIGHT is not a setable attribute for BROWSE "+
                                widgetName());
   }

   /**
    * Implements the ROW-HEIGHT-CHARS attribute getter.<p>
    *
    * @return row height, in characters, of the browse.
    */
   public decimal getRowHeightChars()
   {
      return new decimal(config.rowHeightChars);
   }

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute getter.<p>
    *
    * @return <code>true</code> if column searching is allowed for the browse.
    */
   public logical isAllowColumnSearching()
   {
      return new logical(config.allowColumnSearching);
   }

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   public void setAllowColumnSearching(boolean allow)
   {
      setAllowColumnSearching(new logical(allow));
   }

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   public void setAllowColumnSearching(logical allow)
   {
      if (allow.isUnknown())
         return;

      if (config.allowColumnSearching == allow.booleanValue())
      {
         return;
      }
      
      config.allowColumnSearching = allow.booleanValue();
      pushWidgetAttr("allowColumnSearching", config.allowColumnSearching);
   }

   /**
    * Implements the SEPARATORS attribute getter.<p>
    *
    * @return <code>true</code> if row and column separators of the browse widget are displayed.
    */
   public logical isSeparators()
   {
      return new logical(config.isSeparators());
   }

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   public void setMaxDataGuess(double maxDataGuess)
   {
      setMaxDataGuess(new decimal(maxDataGuess));
   }

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   public void setMaxDataGuess(NumberType maxDataGuess)
   {
      if (maxDataGuess.isUnknown())
         return;

      if (config.maxDataGuess == maxDataGuess.intValue())
      {
         return;
      }
      
      config.maxDataGuess = maxDataGuess.intValue();
      pushWidgetAttr("maxDataGuess", config.maxDataGuess);
   }

   /**
    * Get value of the MAX-DATA-GUESS attribute.
    *
    * @return   the estimated number of records in a browse query.
    */
   public integer getMaxDataGuess()
   {
      return new integer(config.maxDataGuess);
   }
   
   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the repositioned row number for the
    * browse.
    *
    * @param    row
    *           1-based row number where the new record is displayed
    *
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @Override
   public logical setRepositionedRow(int64 row)
   {
      return setRepositionedRow(row, new character());
   }
   
   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the repositioned row number for the
    * browse.
    *
    * @param    row
    *           1-based row number where the new record is displayed
    *
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @Override
   public logical setRepositionedRow(long row)
   {
      return setRepositionedRow(new int64(row), new character());
   }

   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    *
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    conditional
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    *
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @Override
   public logical setRepositionedRow(long row, String conditional)
   {
      return setRepositionedRow(new int64(row), new character(conditional));
   }

   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    *
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    conditional
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    *
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @Override
   public logical setRepositionedRow(int64 row, String conditional)
   {
      return setRepositionedRow(row, new character(conditional));
   }

   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    *
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    conditional
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    *
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   public logical setRepositionedRow(int64 row, character conditional)
   {
      Boolean cond = null;
      if (!conditional.isUnknown())
      {
         String val = conditional.getValue();
         if (val.equalsIgnoreCase("CONDITIONAL"))
         {
            cond = true;
         }
         else if (val.equalsIgnoreCase("ALWAYS") || val.isEmpty())
         {
            cond = false;
         }
         else
         {
            ErrorManager.recordOrShowError(4065,
                  String.format("**The SET-REPOSITIONED-ROW attribute on the BROWSE %s has " +
                        "invalid arguments", name().toStringMessage()),
                  false,
                  false);
            return new logical(false);
         }
      }

      return setRepositionedRow(row.isUnknown() ? 0 : row.intValue(), cond);
   }

   /**
    * Implements the SET-REPOSITIONED-ROW() widget method.
    *
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    conditional
    *           If <code>true</code>, tries to reposition to the browse viewport existing row
    *           first. If <code>false</code> it always uses the indicated row number.
    *           <code>null</code> preserves the current value.
    *
    * @return   Always <code>true</code>.
    */
   public logical setRepositionedRow(int row, final Boolean conditional)
   {
      int down = getDown().intValue();
      if (row < 1)
      {
         row = 1;
      }
      else if (row > down)
      {
         row = down;
      }

      final int r = row - 1;

      config.repositionedRow = r;
      if (conditional != null)
         config.conditionalReposition = conditional;

      pushWidgetAttr(new String[] {"repositionedRow", "conditionalReposition"},
                     new Object[] {config.repositionedRow, config.conditionalReposition});

      return new logical(true);
   }
   
   /**
    * Blocks attempts to set the DCOLOR attribute as it is read-only.
    * 
    * @param   d
    *          Flag is not used.
    */
   public void setDColor(int64 d)
   {
      throw new RuntimeException("DCOLOR attribute is read only");
   }

   /**
    * Blocks attempts to get the PFCOLOR attribute as it is undefined.
    * 
    * @return  nothing 
    */
   public integer getPfColor()
   {
      throw new RuntimeException("PFCOLOR attribute is undefined");
   }

   /**
    * Blocks attempts to set the PFCOLOR attribute as it is undefined.
    * 
    * @param   p
    *          Flag is not used.
    */
   public void setPfColor(int64 p)
   {
      throw new RuntimeException("PFCOLOR attribute is undefined");
   }

   /**
    * Set the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setTitleFont(int64 fontNum)
   {
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized

      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.titleFont = font;
      pushWidgetAttr("titleFont", config.titleFont);
   }

   /**
    * Get the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getTitleFont()
   {
      int font = config.titleFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Creates an entry in the result list for the current row. Implements the 4GL
    * CREATE-RESULT-LIST-ENTRY() method. Created entry contains id(s) of the record(s) currently
    * located in the backing buffer(s).
    *
    * @return <code>true</code> on success.
    */
   @Override
   public logical createResultListEntry()
   {
      if (query == null || !query.isOpen().booleanValue() || !isNewRow().booleanValue())
      {
         ErrorManager.recordOrThrowError(9483,
            "Selected row must be new row when using CREATE-RESULT-LIST-ENTRY for BROWSE widget");
         return new logical(false);
      }

      logical res = query.createResultListEntryWorker();
      if (res.booleanValue())
      {
         LogicalTerminal.rowEntryCreated(this, query.currentRowImpl().intValue() - 1);
      }

      return res;
   }

   /**
    * This will delete the current row in the query's results.  This implements the
    * 4GL DELETE-RESULT-LIST-ENTRY() method.
    *
    * @return  <code>true</code> on success.
    */
   @Override
   public logical deleteResultListEntry()
   {
      try
      {
         LogicalTerminal.checkDeleteEntryRowIndex(this);
         int deletedRowId = query.currentRowImpl().intValue() - 1;
         logical res = query.deleteResultListEntryWorker();
         if (res.booleanValue())
         {
            LogicalTerminal.rowEntryDeleted(this, deletedRowId);
         }
         return res;
      }
      catch (ErrorConditionException e)
      {
         handleError(e, "DELETE-RESULT-LIST-ENTRY");
         return new logical(false);
      }
   }

   // ------------------------------------------------------------------
   // Client side BrowseModel request handlers
   // ------------------------------------------------------------------
   /**
    * Get configurations of editors (fill-ins, toggle-boxes or combo-boxes) used for in-browse
    * editing.
    *
    * @return   Array of configurations of editor widgets (fill-ins, toggle-boxes or combo-boxes)
    *           which are used for in-browse editing. An array element is configuration of the
    *           editor widget for the corresponding editable column or <code>null</code> if the
    *           column is not editable. Return value is <code>null</code> if the browse is
    *           read-only.
    */
   public WidgetConfig[] getEditorsConfigs()
   {
      if (!isEditPossible())
         return null;

      int size = columns.size();
      WidgetConfig[] res = new WidgetConfig[size];
      for (int i = 0; i < size; i++)
      {
         GenericWidget editor = columns.get(i).getEditor();
         res[i] = editor == null ? null : editor.config;
      }

      return res;
   }

   /**
    * Get the number of rows iterated by the query (for preselect queries - matches result size,
    * for adaptive queries - matches the size of the part of the result set the query has already
    * iterated over).
    *
    * @return  number of iterated rows in the result set.
    */
   public int getRowCount()
   {
      if (query == null)
      {
         return 0;
      }

      integer res = query.size();
      if (res.isUnknown())
         return 0;

      return res.intValue();
   }

   /**
    * Get number of available rows starting from the given one.
    * TODO May only be used for preselect queries in GUI. Do not use before adapting it to that
    * use case.
    *
    * @param   startRow
    *          Initial row to start count rows from.
    *
    * @return  Number of available rows or 0 if no query specified.
    */
   public int getRowCount(int startRow)
   {
      if (query == null)
      {
         return 0;
      }

      int res = 0;

      integer rowId = query.currentRowImpl();
      boolean isFoR = query.isFetchOnReposition();
      OffEnd offEnd = query.getOffEnd();

      try
      {
         ignoreRepo++;

         query.setFetchOnReposition(false);

         if (query.isNativelyPreselect())
         {
            query.first();
            res = query.size().intValue();
         }
         else
         {
            query.forward(Browse.CACHE_SIZE * 5);
            res = getCurrentRow();
            if (res < 0)
               res = 0;
         }
      }
      catch (QueryOffEndException exc)
      {
         // Thrown if query.first() cannot move to a first result because the
         // result set is empty.
      }
      finally
      {
         ignoreRepo--;
         query.setFetchOnReposition(isFoR);

         if (!rowId.isUnknown())
         {
            reposition(rowId, true);
         }
         else if (offEnd == OffEnd.BACK)
         {
            query.last();
         }
         else if (res > 0)
         {
            query.first();
         }
      }

      return res;
   }

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    startRowIndex
    *           Index of the first row to retrieve (0-based).
    * @param    count
    *           Maximum count of rows to return. <code>-1</code> for no limit.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction,
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    * @param    displayStrategy
    *           Specifies type of action for which rows are fetched and defines how rows are
    *           iterated / displayed.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public BrowseRow[] getRows(int startRowIndex,
                              int count,
                              boolean forward,
                              int rowDisplayTriggerId,
                              ScreenBuffer[] sb,
                              RowDisplayStrategy displayStrategy)
   {
      return getRows(startRowIndex,
                     count,
                     forward,
                     null,
                     0,
                     -1,
                     -1,
                     -1,
                     rowDisplayTriggerId,
                     sb,
                     displayStrategy);
   }

   /**
    * This function is used when search by key is performed in a browse. Retrieves the set of
    * rows which should be in view after search is completed. There are complex rules which
    * determine the top row of the view depending on how the <i>current</i> row is positioned
    * relatively to the current view and how the <i>found</i> row is positioned relatively to the
    * current view and to the current row.
    *
    * @param    searchChar
    *           Character to be searched for. Search is performed using only the first character
    *           (in the visual data representation) in the current column. Specifying a character
    *           puts this function in the search mode. This parameter is <code>null</code> for
    *           non-search mode.
    * @param    searchColumn
    *           For search mode: 0-based index of the column to be searched.
    * @param    viewportTopRow
    *           For search mode: 0-based index (in the result set) of the top row in the current
    *           view.
    * @param    viewportBottomRow
    *           For search mode: 0-based index (in the result set) of the bottom row in the
    *           current view.
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public BrowseRow[] getRows(char searchChar,
                              int searchColumn,
                              int viewportTopRow,
                              int viewportBottomRow,
                              int rowDisplayTriggerId,
                              ScreenBuffer[] sb)
   {
      return getRows(viewportBottomRow + 1,
                     -1,
                     true,
                     searchChar,
                     searchColumn,
                     -1,
                     viewportTopRow,
                     viewportBottomRow,
                     rowDisplayTriggerId,
                     sb,
                     SEARCH_BY_KEY);
   }

   /**
    * Depending on whether search by key is performed (<code>searchChar</code> parameter is
    * specified):
    * <ul>
    *    <li>In non-search mode: retrieve several subsequent rows from the database.</li>
    *    <li>In search mode: retrieve the set of rows which should be in view after search is
    *    completed. There are complex rules which determine the top row of the view depending
    *    on how the <i>current</i> row is positioned relatively to the current view and how the
    *    <i>found</i> row is positioned relatively to the current view and to the current
    *    row.</li>
    * </ul>
    *
    * @param    startRowIndex
    *           Index of the first row to retrieve (0-based).
    * @param    count
    *           Maximum count of rows to return. <code>-1</code> for no limit.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction,
    * @param    searchChar
    *           Character to be searched for. Search is performed using only the first character
    *           (in the visual data representation) in the current column. Specifying a character
    *           puts this function in the search mode. This parameter is <code>null</code> for
    *           non-search mode.
    * @param    searchColumn
    *           For search mode: 0-based index of the column to be searched.
    * @param    searchFinalRow
    *           For search mode: 0-based index of the row in the result set on which the search
    *           should be stopped (the final row itself is included in the searchable set).
    * @param    viewportTopRow
    *           For search mode: 0-based index (in the result set) of the top row in the current
    *           view.
    * @param    viewportBottomRow
    *           For search mode: 0-based index (in the result set) of the bottom row in the
    *           current view.
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    * @param    displayStrategy
    *           Specifies type of action for which rows are fetched and defines how rows are
    *           iterated / displayed.
    * 
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public BrowseRow[] getRows(int startRowIndex,
                              int count,
                              boolean forward,
                              Character searchChar,
                              int searchColumn,
                              int searchFinalRow,
                              int viewportTopRow,
                              int viewportBottomRow,
                              int rowDisplayTriggerId,
                              ScreenBuffer[] sb,
                              RowDisplayStrategy displayStrategy)
   {
      if (query == null || !query.isOpen().booleanValue())
      {
         return null;
      }

      if (count == 0)
      {
         return new BrowseRow[0];
      }

      boolean searchMode = searchChar != null;
      boolean rowFound   = false;
      boolean chui = LogicalTerminal.isChui();

      // TRUE if we've searched to the end of result set and continued the search from the
      // beginning of the result set.
      boolean cycle      = searchFinalRow >= 0;

      ArrayList<BrowseRow> data = count > 0 ? new ArrayList<>(count) : new ArrayList<>();
      if (!chui &&
          (displayStrategy == INCREMENTAL_SCROLL_FWD ||
           displayStrategy == INCREMENTAL_SCROLL_BACK))
      {
         intermediateRows = data;
         LogicalTerminal.setRowDisplayDrawListener(this);
      }

      integer currentRowId = query.currentRowImpl();
      boolean fOr = query.isFetchOnReposition();

      try
      {
         ignoreRepo++;

         integer repositionRow = new integer(startRowIndex + (forward ? 1 : 2));
         reposition(repositionRow, false);
         if (!forward && CompareOps._isLessThan(query.currentRowImpl(), repositionRow))
         {
            // reposition() works in the way that the following next() call will get the target
            // record #2. That means that the query may be positioned on the previous record #1 or
            // between the records and #1 and #2. But for previous() calls in which we expect to
            // get record #1 we have to ensure that the query is positioned between the records
            // and #1 and #2 or on the record #2.
            query.next();
         }
         query.setFetchOnReposition(true);
         initMaxDataGuess();

         if (count < 0)
         {
            count = Integer.MAX_VALUE; // search until row is found or we've reached the end
         }

         for (int i = 0; i < count; i++)
         {
            if (forward)
            {
               query.next();
            }
            else
            {
               query.previous();
            }

            if (query._isOffEnd())
            {
               break;
            }

            // Not off-end, but the first buffer is not available, most likely it is
            // DELETE-CURRENT-ROW case.
            handle bufferHandle = query.bufferHandle(1);
            if (!bufferHandle.unwrapBuffer()._available())
            {
               break;
            }

            int size = columns.size();

            integer row = query.currentRowImpl();
            if (row.isUnknown())
            {
               StringBuilder errorString = new StringBuilder("Cannot get number of the row [");
               BrowseRow browseRow = new BrowseRow(size);
               fillRowData(browseRow);
               for (int j = 0; j < size; j++)
               {
                  if (j != 0)
                  {
                     errorString.append(", ");
                  }
                  errorString.append(browseRow.data[j].toStringExport());
               }
               errorString.append("]");

               throw new IllegalStateException(errorString.toString());
            }

            if (row.intValue() > maximumKnownRowIndex)
            {
               maximumKnownRowIndex = row.intValue();
            }

            BrowseRow browseRow = new BrowseRow(size, row.intValue() - 1);
            try
            {
               // Fire triggers.
               // On initial row display browse gets DOWN + 1 rows to get the idea if the number
               // of rows is more than DOWN. The extra row is not rendered.
               if (rowDisplayTriggerId != -1 &&
                   !(FIRST_SCROLL.equals(displayStrategy) && data.size() + 1 == count))
               {
                  try
                  {
                     // rowInDisplay holds screen values
                     rowInDisplay = new BrowseRow(size, row.intValue() - 1);
                     LogicalTerminal.trigger(rowDisplayTriggerId,
                                             Keyboard.eventCode(Keyboard.ROW_DISPLAY),
                                             getId(),
                                             -1,
                                             -1,
                                             -1,
                                             sb);
                  }
                  finally
                  {
                     fillRowData(browseRow);
                     rowInDisplay = null;
                  }
               }
               else
               {
                  fillRowData(browseRow);
               }

               browseRow.renderRow(getColumnConfigs());
            }
            finally
            {
               // current row could be adjusted by trigger or by error messages because of focus
               // changes: when browse gains the focus back, it sets the current (highlighted, not
               // triggered) browse row as the current query row.
               if (!row.equals(query.currentRowImpl()))
               {
                  query.reposition(row);
               }
            }

            data.add(browseRow);

            if (searchMode)
            {
               // trim to DOWN size
               if (data.size() > config.down)
               {
                  data.remove(0);
               }

               if (Browse.cellStartsWith(browseRow.text[searchColumn], searchChar))
               {
                  currentRowId = query.currentRowImpl();
                  rowFound = true;
                  break;
               }

               if (cycle && row.intValue() > searchFinalRow)
               {
                  break;
               }
            }
         }

         // update maxDataGuess
         if (forward)
         {
            if (maximumKnownRowIndex > getMaxDataGuess().intValue() ||
                data.size() < count)
            {
               // We have exceeded current data guess or we have reached the last row
               realDataGuessUsed = true;
            }
         }
      }
      catch (QueryOffEndException exc)
      {
         // potentially, may be thrown in reposition() because the result set
         // is empty.
      }
      finally
      {
         if (!chui)
         {
            intermediateRows = null;
            LogicalTerminal.setRowDisplayDrawListener(null);
         }
         ignoreRepo--;

         try
         {
            // Now reposition to where we were originally.
            if (!currentRowId.isUnknown())
               reposition(currentRowId, true);
            else
               query.first();
         }
         catch (QueryOffEndException exc)
         {
            // thrown in reposition() if the result set is empty.
         }
         finally
         {
            query.setFetchOnReposition(fOr);
         }
      }

      if (!forward)
      {
         Collections.reverse(data);
      }

      BrowseRow[] result = data.toArray(new BrowseRow[data.size()]);

      if (searchMode)
      {
         if (rowFound)
         {
            int foundRow = result[result.length - 1].rowIndex;

            if (!cycle)
            {
               // Before cycle

               // Structure or rows:
               //                        =              =
               //upper part (optional)   =              =
               //                        = current row  =
               //                        ================
               //                        =              =
               //                        =              =
               //                        =              =
               //                        = found row    =
               //                        ================
               // lower part (optional)  =              =
               //                        =              =

               int rowsToFetch = config.down - result.length;
               if (rowsToFetch > 0)
               {
                  int divisionRow = result[0].rowIndex;
                  if (divisionRow > 0)
                  {
                     BrowseRow[] upperPart = null;
                     if (isRowOnFirstPage(foundRow))
                     {
                        // found row is within the first page: add the "current row",
                        // i.e. (startRow - 1)
                        if (startRowIndex > 0)
                        {
                           upperPart = getRows(startRowIndex - 1,
                                               1,
                                               true,
                                               rowDisplayTriggerId,
                                               sb,
                                               displayStrategy);
                           if (upperPart.length > 0 &&
                               result[0].rowIndex != startRowIndex) // 1-based vs
                                                                    // 0-based
                           {
                              // "current row" was deleted, drop upper part
                              upperPart = null;
                           }
                        }
                     }
                     else
                     {
                        // fetch upper part
                        upperPart = getRows(divisionRow - 1,
                                            rowsToFetch,
                                            false,
                                            rowDisplayTriggerId,
                                            sb,
                                            displayStrategy);
                     }

                     if (upperPart != null && upperPart.length > 0)
                     {
                        BrowseRow[] newResult = new BrowseRow[upperPart.length + result.length];
                        System.arraycopy(upperPart, 0, newResult, 0, upperPart.length);
                        System.arraycopy(result, 0, newResult, upperPart.length, result.length);
                        result = newResult;
                     }
                  }

                  // still not enough rows to fill the view?
                  result = addRowsToBottom(result, rowDisplayTriggerId, sb, displayStrategy);
               }

               if (maximumKnownRowIndex > getMaxDataGuess().intValue())
               {
                  // push max data guess to client
                  setMaxDataGuess(maximumKnownRowIndex);
               }
            }
            else
            {
               // After cycle

               if (!isRowInView(viewportTopRow, viewportBottomRow, foundRow))
               {
                  // found row is out the view - found row is the top row, leave only it and add
                  // rows to bottom
                  BrowseRow[] newResult = new BrowseRow[1];
                  System.arraycopy(result, result.length - 1, newResult, 0, 1);
                  result = newResult;
               }

               result = addRowsToBottom(result, rowDisplayTriggerId, sb, displayStrategy);
            }
         }
         else if (!cycle)
         {
            // push max data guess to client
            config.maxDataReached = true;
            setMaxDataGuess(maximumKnownRowIndex);

            // cycle to find row from top
            return getRows(0,
                           -1,
                           true,
                           searchChar,
                           searchColumn,
                           Math.max(startRowIndex - 1, 0),
                           viewportTopRow,
                           viewportBottomRow,
                           rowDisplayTriggerId,
                           sb,
                           displayStrategy);
         }
         else
         {
            // we've searched through the whole result set and didn't find a row
            if (currentRowId.isUnknown())
            {
               // empty result set or some current row sync error
               return getRows(0, config.down, true, rowDisplayTriggerId, sb, displayStrategy);
            }

            int currentRow = currentRowId.intValue() - 1;
            if (viewportTopRow >= 0 && viewportTopRow >= currentRow)
            {
               // current row is before the view - page up (by DOWN - 1)
               return getRows(Math.max(0, viewportTopRow - (config.down - 1)),
                              config.down,
                              true,
                              rowDisplayTriggerId,
                              sb,
                              displayStrategy);
            }
            else if (isRowInView(viewportTopRow, viewportBottomRow, currentRow))
            {
               // current row is in the view
               return addRowsToBottom(result, rowDisplayTriggerId, sb, displayStrategy);
            }
            else
            {
               // current row is after the view - restore the same view
               return getRows(viewportTopRow,
                              config.down,
                              true,
                              rowDisplayTriggerId,
                              sb,
                              displayStrategy);
            }
         }

         // TODO max data guess in 4GL is not updated right away:
         // 1. at first thumb takes incorrect position (thumb size doesn't change) - in FWD this
         // step is skipped;
         // 2. after an action is taken (mouse click or key press), thumb takes correct size
         // (taking into consideration rows that were fetched during search) and position.
      }

      return result;
   }

   /**
    * Fills the current row container with row data taking into account that the cell screen value
    * can override the column value and its getter is not evaluated in this case.
    * 
    * @param    browseRow
    *           The current row container
    * 
    * @return   The current row container
    */
   private BrowseRow fillRowData(BrowseRow browseRow)
   {
      int size = browseRow.data.length;
       // fill row data
       for (int j = 0; j < size; j++)
       {
          BrowseColumnWidget column = columns.get(j);
          
          BaseDataType val;
          
          // override value using screen value
          if (rowInDisplay != null && rowInDisplay.data[j] != null)
          {
             val = rowInDisplay.data[j];
          }
          else
          {
             val = column.getValue();
          }

          if (!column.isCalcColumn())
          {
             fixDataTypeOverride(column, val);
          }
          
          browseRow.data[j] = val;

          if (rowInDisplay != null)
          {
             browseRow.attr[j] = rowInDisplay.attr[j];
          }
       }
       
       return browseRow;
   }

   /**
    * Push "intermediate" rows (i.e. the rows that were already fetched in the process of getting
    * new rows by {@link #getRows(int, int, boolean, int, ScreenBuffer[], RowDisplayStrategy)})
    * to the client, so these rows can be drawn if a pause occurs in ROW-DISPLAY trigger.
    */
   @Override
   public void pushIntermediateRows()
   {
      if (intermediateRows != null && intermediateRows.size() > 0)
      {
         BrowseRow[] rows = intermediateRows.toArray(new BrowseRow[intermediateRows.size()]);
         LogicalTerminal.pushIntermediateRows(this, rows);
      }
   }

   /**
    * Updates selected cells in the given row or nothing to do if no query
    * specified.
    * 
    * @param    rowIndex
    *           0-based index of the row to update.
    * @param    cellIndices
    *           Cells for update indexes.
    * @param    cellValues
    *           Cells for update new values.
    * @param    onlyNotifyRowReleased
    *           Only notify server that the row has been released. Data is not updated.
    * 
    * @throws   DataValidationException
    *           New values did not pass database validation rules.
    */
   public void updateRow(int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues,
                         boolean        onlyNotifyRowReleased)
   throws DataValidationException
   {
      if (query == null)
      {
         return;
      }

      if (!onlyNotifyRowReleased)
      {
         integer rowId = query.currentRowImpl();
         String msg = null;

         reposition(new integer(rowIndex + 1), true);

         BrowseColumnWidget[] c = new BrowseColumnWidget[columns.size()];

         c = columns.toArray(c);

         TransactionManager.pushScope("update",
               TransactionManager.TRANSACTION,
               false,
               false,
               false,
               false);
         try
         {
            update:
            {
               TransactionManager.blockSetup();

               do
               {
                  try
                  {
                     for (int i = 0; i < cellIndices.length; i++)
                     {
                        c[cellIndices[i]].setValue(cellValues[i]);
                     }
                  }
                  catch (ErrorConditionException err_except)
                  {
                     TransactionManager.honorError(err_except);
                     TransactionManager.rollback("update");
                     TransactionManager.triggerRetry(null);
                     msg = err_except.getMessage();
                  }
                  catch (RetryUnwindException ru_except)
                  {
                     TransactionManager.honorRetry(ru_except);
                  }
               }
               while (TransactionManager.needsRetry());

               if (TransactionManager.isBreakPending())
               {
                  break update;
               }
            }
         }
         finally
         {
            TransactionManager.popScope();
         }

         reposition(rowId, true);

         if (msg != null)
            throw new DataValidationException(msg);
      }

      WidgetConfig[] editors  = getEditorsConfigs();
      if (editors != null)
      {
         for (WidgetConfig cfg : editors)
         {
            if (cfg != null)
            {
               cfg.modified = false;
            }
         }
      }
   }

   /**
    * Returns <code>true</code> if specified row is new and is not present
    * in the data base.
    * 
    * @param    rowIndex
    *           Index of the row to return information about.
    *          
    * @return   <code>true</code> if row is new and <code>false</code>
    *           otherwise.
    */
   public boolean isNewRow(int rowIndex)
   {
      if (query == null)
      {
         return false;
      }

      integer save = query.currentRowImpl();
      
      reposition(new integer(rowIndex + 1), false);

      boolean res = query.isOffEnd().booleanValue(); 
      
      reposition(save, false);

      return res;
   }
   
   /**
    * Set new current row in the browse. This method is called as notification from the client
    * that current row is changed.
    * 
    * @param   rowIndex
    *          0-based index of the current row. If negative, then the row is virtual so the
    *          buffers will be emptied. 
    * @param   force
    *          Force positioning to the specified row, even if it is selected.
    *          
    * @return  New current row or 0 if no query specified.
    */
   public int setCurrentRow(int rowIndex, boolean force)
   {
      if (query == null || !query.isOpen().booleanValue())
      {
         return 0;
      }
      else if (rowIndex < 0)
      {
         // the browse requested selection of a virtual row. We need to empty the buffers now
         query.releaseBuffers(); // The buffers are released when new row entry event is triggered.
         
         return 0;
      }
      
      if (!force && rowIndex == getCurrentRow())
      {
         return rowIndex;
      }
      
      ignoreRepo++;
      boolean fOr = query.isFetchOnReposition();
      
      try
      {
         query.setFetchOnReposition(true);
         reposition(new integer(rowIndex + 1), true);
         logical offEnd = query.isOffEnd();
         if (!offEnd.isUnknown() && offEnd.booleanValue())
         {
            // If we've gone too far, reset ourselves to the last record.
            query.last();
         }
      }
      catch (QueryOffEndException exc)
      {
         // Thrown if result set is empty.
      }
      finally
      {
         ignoreRepo--;
         query.setFetchOnReposition(fOr);
      }
      
      return getCurrentRow();
   }

   /**
    * Get current row in the query.
    *
    * @return   The current row index (0-based) or -1 if the query is off-end.
    */
   public int getCurrentRow()
   {
      if (query == null)
      {
         return -1;
      }

      integer row = query.currentRow();
      return (row.isUnknown() ? -1 : row.intValue() - 1);
   }
   
   //---------------------------------------------------------------
   
   /**
    * Set new value for the DOWN attribute.
    *
    * @param    down
    *           New value for the attribute.
    */
   public void setDown(final long down)
   {
      config.down = (int) down;
      pushWidgetAttr("down", config.down);
   }

   /**
    * Java implementation for setting the DOWN attribute which contains number of iterations in
    * a down frame or number of potential rows in a browse widget.
    *
    * @param down Value of the DOWN attribute.
    */
   public void setDown(int64 down)
   {
      if (down.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute DOWN on BROWSE " + widgetName(),
               false);

         return;
      }

      setDown(down.intValue());
   }

   /**
    * Get the vale of the DOWN attribute.
    *
    * @return value of the DOWN attribute.
    */
   public integer getDown()
   {
      return new integer(config.down);
   }

   /**
    * Add new field to the list of browse fields.
    * 
    * @param    column
    *           Field to add.
    */
   public void addColumn(BrowseColumnWidget column)
   {
      addColumn(column, -1);
   }

   /**
    * Add new field to the list of browse fields.
    *
    * @param    column
    *           Field to add.
    * @param    position
    *           0-based position of the column, <code>-1</code> to add it as the last column.
    */
   public void addColumn(BrowseColumnWidget column, int position)
   {
      if (position < 0 || position >= columns.size())
      {
         // add column as last
         column.config().ordinal = columns.size();
         columns.add(column);
      }
      else
      {
         column.config().ordinal = position;
         columns.add(position, column);
         int size = columns.size();
         for (int i = position + 1; i < size; i++)
            columns.get(i).config().ordinal = i;
      }

      column.setBrowseId(getId());
   }

   /**
    * Change editable state for all columns.
    * 
    * @param    editable
    *           New value for the attribute.
    */
   public void setEditableAll(boolean editable)
   {
      for (BrowseColumnWidget column : columns)
      {
         column.setEditable(editable);
      }
   }
   
   /**
    * Overrides the parent version to notify that this is not a queryable attribute.
    *
    * @return   Always returns the unknown value.
    */
   @Override
   public character getScreenValue()
   {
      notQueryable("SCREEN-VALUE");
      return new character();
   }
   
   /**
    * Overrides the parent version to notify that this is not a settable attribute.
    *
    * @param    value
    *           Ignored.
    */
   @Override
   public void setScreenValue(character value)   
   {
      notSettable("SCREEN-VALUE");
   }

   /**
    * Gets the SCROLLABLE writeable attribute.
    *
    * @return   The current value of the SCROLLABLE attribute.
    */
   @Override
   public logical isScrollable()
   {
      return new logical(config.scrollable);
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   @Override
   public void setScrollable(logical scroll)
   {
      if (scroll.isUnknown())
         return;

      setScrollable(scroll.booleanValue());
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   @Override
   public void setScrollable(boolean scroll)
   {
      if (config.scrollable == scroll)
      {
         return;
      }

      config.scrollable = scroll;
      pushWidgetAttr("scrollable", config.scrollable);
   }

   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   @Override
   public void setBgColor(int64 bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setBgColor(color);
   }

   /**
    * Get the SEPARATOR-FGCOLOR attribute value.
    *
    * @return SEPARATOR-FGCOLOR attribute value.
    */
   public integer getSeparatorFgColor()
   {
      return config.separatorFgColor == -1 ? new integer() : new integer(config.separatorFgColor);
   }

   /**
    * Get the COLUMN-FGCOLOR attribute value.
    *
    * @return COLUMN-FGCOLOR attribute value.
    */
   public integer getColumnFgColor()
   {
      return config.columnFgColor == -1 ? new integer() : new integer(config.columnFgColor);
   }


   /**
    * Get the COLUMN-BGCOLOR attribute value.
    *
    * @return COLUMN-BGCOLOR attribute value.
    */
   public integer getColumnBgColor()
   {
      return config.columnBgColor == -1 ? new integer() : new integer(config.columnBgColor);
   }

   /**
    * Sets the SEPARATOR-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   public void setSeparatorFgColor(long color)
   {
      if (config.separatorFgColor == color)
      {
         return;
      }
      
      config.separatorFgColor = (int) color;
      pushWidgetAttr("separatorFgColor", config.separatorFgColor);
   }

   /**
    * Sets the SEPARATOR-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   public void setSeparatorFgColor(int64 color)
   {
      setSeparatorFgColor(color.isUnknown() ? -1 : color.intValue());
   }

   /**
    * Sets the COLUMN-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   public void setColumnFgColor(long color)
   {
      if (config.columnFgColor == color)
      {
         return;
      }
      
      config.columnFgColor = (int) color;
      pushWidgetAttr("columnFgColor", config.columnFgColor);
   }

   /**
    * Sets the COLUMN-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   public void setColumnFgColor(int64 color)
   {
      setColumnFgColor(color.isUnknown() ? -1 : color.intValue());
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   public void setColumnBgColor(long color)
   {
      if (config.columnBgColor == color)
      {
         return;
      }
      
      config.columnBgColor = (int) color;
      pushWidgetAttr("columnBgColor", config.columnBgColor);
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   public void setColumnBgColor(int64 color)
   {
      setColumnBgColor(color.isUnknown() ? -1 : color.intValue());
   }

   /**
    * Register query and list of fields with the browse.
    * 
    * @param    query
    *           Reference to query. May be null.
    * 
    * @param    list
    *           List of fields. May be null.
    */
   public void registerQuery(P2JQuery query, FrameElement[] list)
   {
      if (query != null)
      {
         setQuery(query);
      }

      if (list != null)
      {
         for (int i = 0; i < list.length; i++)
         {
            // provide link
            ((BrowseColumnWidget) list[i].getWidget()).setElement(list[i]);
         }
      }
   }

   /**
    * Finish initialization of the browse widget. 
    */
   public void finishSetup()
   {
      if (config.applyEnhancedLayout && !enhancedInitDone && query != null)
      {
         enhancedInit();
      }

      if (initDone)
         return;
    
      if (query == null)
      {
         initDone = true;
         return;
      }

      createEditors();
      
      // finish initialization of the widget
      try
      {
         // retrieve the first record only if the query is already opened
         if (query.isOpen().booleanValue() && query.currentRowImpl().isUnknown())
            query.first();
      }
      catch (QueryOffEndException exc)
      {
         // OK to ignore;  indicates query has no results and browse will be
         // empty initially.
      }
      config.name = widgetName();
      pushScreenDefinition();

      initDone = true;
   }

   /**
    * Cleanup dynamic browse before deletion.
    */
   public void cleanup()
   {
      if (query != null)
      {
         query.setBrowsed(false);
         query.removeRepositionListener(this);
      }

      if (frame != null)
      {
         Iterator<BrowseColumnWidget> iter = getBrowseColumns();
         while (iter.hasNext())
         {
            BrowseColumnWidget column = iter.next();
            frame.deleteDynamicWidget(column, false);
            iter.remove();
         }
      }
   }

   /**
    * Invoked when the query with which this browse widget is associated is
    * explicitly repositioned.
    *
    * @param error
    *        <code>true</code> if an error happened during reposition.
    * @param targetRepositionRow
    *        Row used as the target for reposition. If it was deleted, reposition with fetching
    *        may end up on a different row.
    */
   public void queryRepositioned(boolean error, int targetRepositionRow)
   {
      if (ignoreRepo > 0)
      {
         return;
      }

      if (!isRefreshable().booleanValue())
      {
         switch (pendingReopen)
         {
            case NO_REOPEN:
               pendingRepositionRow = getCurrentRow();
               break;
            case PRE_REOPEN:
               pendingReopen = PendingOpenState.REOPEN;
               break;
            case PRE_REPOSITIONED_AND_REOPEN:
               pendingReopen = PendingOpenState.REPOSITIONED_AND_REOPEN;
               break;
            case REOPEN:
               pendingRepositionRow = getCurrentRow();
               break;
            case REPOSITIONED_AND_REOPEN:
               pendingReopen = PendingOpenState.REOPEN;
               pendingRepositionRow = getCurrentRow();
               break;
         }
      }
      else
      {
         LogicalTerminal.queryRepositioned(this, error, targetRepositionRow);
      }
   }

   /**
    * Force refreshing of the data in the current row at the client side.
    */
   public void refreshCurrentRow()
   {
      refreshCurrentRow(null);
   }

   /**
    * Force refreshing of the data in the current row at the client side.
    *
    * @param refreshedColumns
    *        The set of columns for which the data is updated (in the current row).
    *        <code>null</code> if all columns should be refreshed.
    */
   public void refreshCurrentRow(BrowseColumnWidget[] refreshedColumns)
   {
      Set<Integer> displayColumnsSet = null;
      if (refreshedColumns != null)
      {
         displayColumnsSet = new HashSet<>();
         for (BrowseColumnWidget column : refreshedColumns)
         {
            displayColumnsSet.add(column.config().ordinal);
         }
      }

      // fill row data
      int size = columns.size();
      BaseDataType[] rowData = new BaseDataType[size];

      int columnsCount = size;
      for (int i = 0; i < size; i++)
      {
         BrowseColumnWidget column = columns.get(i);
         if ((displayColumnsSet != null && displayColumnsSet.contains(column.config().ordinal)) ||
             column.isCalcColumn())
         {
            columnsCount--;
            continue;
         }

         try
         {
            // screen value has more priority over var value here to take
            // most recent column update while browse row display 
            BaseDataType val = column.screenValueToBDT(column.getScreenValue(), false);
            // however if screen value is unknown we need to get the variable value instead
            if (val.isUnknown() || !column.config().inEditMode)
            {
               val = column.getValue();
            }
            rowData[i] = val;
         }
         catch (ErrorConditionException e) // no record available
         {
            columnsCount--;
         }
      }

      if (columnsCount == 0)
      {
         return;
      }

      BrowseRow browseRow = new BrowseRow(-1,  // not used
                                          rowData,
                                          LogicalTerminal.isChui());
      browseRow.renderRow(getColumnConfigs());

      LogicalTerminal.refreshCurrentRow(this, browseRow);
   }
   
   /**
    * Invoked when the query with which this browse widget is associated is
    * explicitly closed.
    */
   public void queryClosed()
   {
      LogicalTerminal.queryClosed(this);
      
      maximumKnownRowIndex = null;
      
      if (!isRefreshable().booleanValue())
      {
         pendingReopen = pendingRepositionRow == -1 ?
                            PendingOpenState.PRE_REOPEN :
                            PendingOpenState.PRE_REPOSITIONED_AND_REOPEN;
      }
   }

   /**
    * Invoked when a query to which the listener is registered is deleted.
    */
   public void queryDeleted()
   {
      query = null;
   }

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   public boolean isRowAvailable(int rowIndex)
   {
      BrowseRow[] data = getRows(rowIndex, 1, true, -1, null, RowDisplayStrategy.NO_DISPLAY);
      return data.length > 0 && rowIndex == data[0].rowIndex;
   }

   /**
    * Run browse report.
    *
    * @param  outputFileName
    *         Name of the output file.
    * @param  outputFormat
    *         Format of the output format.
    *
    * @return <code>true</code> on success.
    */
   public boolean runReport(String outputFileName, MediaType outputFormat)
   {
      handle h = getReport();
      if (h.isUnknown())
      {
         return false;
      }
      
      FwdReport report = h.unwrapFwdReport();
      String title = getTitle().getValue();
      if (title == null || title.trim().isEmpty())
      {
         title = "Browse Report";
      }
      boolean paged = (outputFormat == MediaType.PDF);
      report.setReportParameter("ReportTitle", title);
      
      // the file format was unknown at the time the [dataSource] was set
      report.setReportDesign(ReportFactory.getDefaultReportTemplate(this, paged));
      
      return report.exportReport(new character(outputFileName), outputFormat).getValue();
   }

   /**
    * Apply default 4GL configuration (non-enhanced) for this browse.
    *
    * @return <code>true</code> on success.
    */
   public boolean applyDefault4glConfig()
   {
      return applyEnhancedConfig(default4glState);
   }

   /**
    * Get save targets for an enhanced browse configuration, which are available for the current
    * user.
    *
    * @return see above.
    */
   public EnhancedBrowseSaveTarget[] getEnhancedBrowseSaveTargets()
   {
      return EnhancedBrowseConfigManager.getInstance().getEnhancedBrowseSaveTargets();
   }

   /**
    * Save the current enhanced browse configuration using the specified save target.
    *
    * @param  saveTarget
    *         Save target (target server, target account, applied to all browses or not).
    *
    * @return <code>true</code> on success.
    */
   public boolean saveEnhancedConfig(EnhancedBrowseSaveTarget saveTarget)
   {
      EnhancedBrowseConfig config = new EnhancedBrowseConfig(this);
      boolean res = EnhancedBrowseConfigManager.getInstance().saveEnhancedConfig(config,
                                                                                 saveTarget);
      if (!res)
      {
         LogicalTerminal.messageBox(
               "Can't save enhanced browse configuration, see log for more details.",
               false,
               null,
               LogicalTerminal.ALERT_ERROR,
               LogicalTerminal.BTN_OK,
               "Saving enhanced browse state");
      }

      return res;
   }

   /**
    * Apply the given enhanced browse configuration to this browse.
    *
    * @param  config
    *         Enhanced browse configuration to apply.
    *
    * @return <code>true</code> on success.
    */
   public boolean applyEnhancedConfig(EnhancedBrowseConfig config)
   {
      // browse parameters
      this.config.ehFgColor = config.ehFgColor;
      this.config.ehBgColor = config.ehBgColor;
      this.config.ehAltBgColor = config.ehAltBgColor;
      this.config.ehLabelFgColor = config.ehLabelFgColor;
      this.config.ehLabelBgColor = config.ehLabelBgColor;
      this.config.ehSeparatorsColor = config.ehSeparatorsColor;
      this.config.ehFont = config.ehFont;
      this.config.ehLabelFont = config.ehLabelFont;
      if (config.labelsHeightPixels != null && config.labelsHeightPixels > 0)
      {
         this.config.labelsHeightPixels = config.labelsHeightPixels;
         this.config.labelsHeightNoEnh  = config.labelsHeightPixels;
      }
      if (config.rowSeparators != null)
      {
         this.config.rowSeparators = config.rowSeparators;
      }
      if (config.columnSeparators != null)
      {
         this.config.columnSeparators = config.columnSeparators;
      }
      if (config.rowHeightPixels != null)
      {
         setRowHeightPixels(new integer(config.rowHeightPixels), false);
      }

      // column parameters
      int colNo = columns.size();
      Map<String, BrowseColumnWidget> columnsMap = new HashMap<>(colNo);
      for (BrowseColumnWidget col : columns)
      {
         columnsMap.put(col.config.ehColumnKey, col);
      }

      if (config.columnConfigs != null && config.columnConfigs.size() > 0)
      {
         BrowseColumnWidget[] newColumns = new BrowseColumnWidget[colNo];
         for (Map.Entry<String, EnhancedColumnConfig> entry : config.columnConfigs.entrySet())
         {
            String columnKey = entry.getKey();
            EnhancedColumnConfig columnState = entry.getValue();

            BrowseColumnWidget col = columnsMap.get(columnKey);
            col.config.ehFgColor = columnState.ehFgColor;
            col.config.ehBgColor = columnState.ehBgColor;
            col.config.ehAltBgColor = columnState.ehAltBgColor;
            col.config.ehLabelFgColor = columnState.ehLabelFgColor;
            col.config.ehLabelBgColor = columnState.ehLabelBgColor;
            col.config.ehFont = columnState.ehFont;
            col.config.ehLabelFont = columnState.ehLabelFont;

            if (columnState.widthPixels != null)
            {
               BaseEntity.CharsPixels charsPixels =
                     BaseEntity.prepareSizePixels(columnState.widthPixels,
                                                  SizeAttribute.WIDTH_CHARS,
                                                  this);
               col.setWidthWorker(charsPixels, false);
            }
            if (columnState.visible != null)
            {
               col.config.visible = columnState.visible;
               col.config.hidden = !col.config.visible;
            }

            // ordinals are recalculated on the client side
            newColumns[columnState.ordinal] = col;
         }
         columns = new ArrayList<>(Arrays.asList(newColumns));
      }
      updateColumnIds();
      pushScreenDefinition();

      notifyColumnsUpdated(true);
      LogicalTerminal.applyEnhancedConfigClient(this, config);

      return true;
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   public void setSensitive(boolean s)
   {
      if (frame == null)
      {
         ErrorManager.recordOrShowError(4073,
               "**Unable to process SENSITIVE attribute.  BROWSE widget is not in a frame",
               false, false);
         return;
      }

      super.setSensitive(s);
   }

   /**
    * Sets the WIDTH-CHARS writable attribute.
    *
    * @param    width
    *           The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(NumberType width)
   {
      if (!validateWidth(width))
      {
         return;
      }

      super.setWidthChars(width);
   }

   /**
    * Sets the WIDTH-CHARS writable attribute.
    *
    * @param    width
    *           The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(double width)
   {
      if (!validateWidth(new decimal(width)))
      {
         return;
      }

      super.setWidthChars(width);
   }

   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   public void setHeightChars(NumberType height)
   {
      if (!validateHeight(height))
      {
         return;
      }

      super.setHeightChars(height);
   }

   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   public void setHeightChars(double height)
   {
      if (!validateHeight(new decimal(height)))
      {
         return;
      }

      super.setHeightChars(height);
   }

   /**
    * Get the LABEL-FONT attribute of this browse column widget.
    *
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getLabelFont()
   {
      int font = config.labelFont;

      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the LABEL-FONT attribute of this browse widget.
    *
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setLabelFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      setLabelFont(fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());
   }

   /**
    * Set the LABEL-FONT attribute of this browse widget.
    *
    * @param    font
    *           An entry in the font-table or unknown to refer the default font.
    */
   public void setLabelFont(int font)
   {
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized

      config.labelFont = font;
      pushWidgetAttr("labelFont", config.labelFont);
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Get the current column (a column can be selected by clicking on its header if
    * ALLOW-COLUMN-SEARCHING is <code>true</code>).
    *
    * @return   handle to the current column .
    */
   @Override
   public handle getCurrentColumn()
   {
      for (BrowseColumnWidget col : columns)
      {
         if (col.config().pressed)
         {
            return new handle(col);
         }
      }

      return new handle();
   }

   /**
    * Set the current column (a column can be selected by clicking on its header if
    * ALLOW-COLUMN-SEARCHING is <code>true</code>).
    *
    * @param    columnHandle
    *           Handle to the new current column.
    */
   @Override
   public void setCurrentColumn(handle columnHandle)
   {
      // seems to do nothing in 4GL
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               String initialValue,
                               String label)
   {
      return addCalcColumn(new character(dataType),
                           new character(format),
                           new character(initialValue),
                           new character(label));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               character format,
                               character initialValue,
                               String label)
   {
      return addCalcColumn(new character(dataType),
                           format,
                           initialValue,
                           new character(label));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               character format,
                               String initialValue,
                               String label)
   {
      return addCalcColumn(new character(dataType),
                           format,
                           new character(initialValue),
                           new character(label));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               String initialValue,
                               character label)
   {
      return addCalcColumn(new character(dataType),
                           new character(format),
                           new character(initialValue),
                           label);
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               character initialValue,
                               character label)
   {
      return addCalcColumn(new character(dataType),
                           new character(format),
                           initialValue,
                           label);
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(character dataType,
                               character format,
                               String initialValue,
                               character label)
   {
      return addCalcColumn(dataType,
                           format,
                           new character(initialValue),
                           label);
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(character dataType,
                               character format,
                               character initialValue,
                               String    label)
   {
      return addCalcColumn(dataType,
                           format,
                           initialValue,
                           new character(label));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    * @param  position
    *         The optional integer value position of the browse column. If pos = 2, the column is
    *         the second column. If the position is not specified or the position is invalid, the
    *         new column is added at the end of the columns.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               String initialValue,
                               String label,
                               int position)
   {
      return addCalcColumn(new character(dataType),
                           new character(format),
                           new character(initialValue),
                           new character(label),
                           new integer(position));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    * @param  position
    *         The optional integer value position of the browse column. If pos = 2, the column is
    *         the second column. If the position is not specified or the position is invalid, the
    *         new column is added at the end of the columns.
    * @param  widgetType
    *         An optional character expression specifying the widget type of the column. Valid
    *         values are "COMBO-BOX", "DROP-DOWN", "DROP-DOWN-LIST", "FILL-IN" and "TOGGLE-BOX".
    *         The default value is "FILL-IN". If datatype-exp is not "LOGICAL", specifying
    *         "TOGGLE-BOX" as the widget type generates a run-time error. The widget-type
    *         parameter is ignored in character mode.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               String initialValue,
                               String label,
                               int position,
                               String widgetType)
   {
      return addCalcColumn(new character(dataType),
                           new character(format),
                           new character(initialValue),
                           new character(label),
                           new integer(position),
                           new character(widgetType));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    * @param  position
    *         The optional integer value position of the browse column. If pos = 2, the column is
    *         the second column. If the position is not specified or the position is invalid, the
    *         new column is added at the end of the columns.
    * @param  widgetType
    *         An optional character expression specifying the widget type of the column. Valid
    *         values are "COMBO-BOX", "DROP-DOWN", "DROP-DOWN-LIST", "FILL-IN" and "TOGGLE-BOX".
    *         The default value is "FILL-IN". If datatype-exp is not "LOGICAL", specifying
    *         "TOGGLE-BOX" as the widget type generates a run-time error. The widget-type
    *         parameter is ignored in character mode.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               character format,
                               character initialValue,
                               String label,
                               integer position,
                               String widgetType)
   {
      return addCalcColumn(new character(dataType),
                           format,
                           initialValue,
                           new character(label),
                           position,
                           new character(widgetType));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(character dataType,
                               character format,
                               character initialValue,
                               character label)
   {
      return addCalcColumn(dataType, format, initialValue, label, new integer());
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    * @param  position
    *         The optional integer value position of the browse column. If pos = 2, the column is
    *         the second column. If the position is not specified or the position is invalid, the
    *         new column is added at the end of the columns.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(character dataType,
                               character format,
                               character initialValue,
                               character label,
                               NumberType position)
   {
      return addCalcColumn(dataType, 
                           format, 
                           initialValue, 
                           label, 
                           position, 
                           new character("FILL-IN"));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               BaseDataType initialValue,
                               String label)
   {
      return addCalcColumn(dataType, format, initialValue, label, -1);
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               int initialValue,
                               character label)
   {
      return addCalcColumn(new character(dataType), 
                           new character(format), 
                           new character(new integer(initialValue)), // TODO: correct? 
                           label, 
                           new integer(-1));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    * @param  position
    *         The optional integer value position of the browse column. If pos = 2, the column is
    *         the second column. If the position is not specified or the position is invalid, the
    *         new column is added at the end of the columns.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               int initialValue,
                               String label,
                               int position)
   {
      return addCalcColumn(dataType,
                           format,
                           new integer(initialValue),
                           label,
                           position);
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    * @param  position
    *         The optional integer value position of the browse column. If pos = 2, the column is
    *         the second column. If the position is not specified or the position is invalid, the
    *         new column is added at the end of the columns.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(String dataType,
                               String format,
                               BaseDataType initialValue,
                               String label,
                               int position)
   {
      // 4GL allows non-string values although it is not officially supported and in most
      // cases rise error message
      return addCalcColumn(new character(dataType),
                           new character(format),
                           new character(initialValue.toStringExport()), // correct?
                           new character(label),
                           new integer(position));
   }

   /**
    * Creates a browse column from the specified properties and returns the handle of the new
    * column. This method can be used only after the browse's query attribute has been set.
    * Besides its name, "calc" column do not calculate anything, its content can be set in
    * ROW-DISPLAY trigger.
    *
    * @param  dataType
    *         Character expression specifying the data type. Valid values are "CHARACTER", "DATE",
    *         "DECIMAL", "INTEGER", "INT64", or "LOGICAL".
    * @param  format
    *         Character expression specifying the column's format.
    * @param  initialValue
    *         Character expression specifying the initial value. This may be a null string.
    * @param  label
    *         Character expression specifying the column's label.
    * @param  position
    *         1-based position of the browse column (i.e. if pos = 2, the column is the second
    *         column). If the position is <code>null</code> or invalid, the new column is added at
    *         the end of the columns.
    * @param  widgetType
    *         An optional character expression specifying the widget type of the column. Valid
    *         values are "COMBO-BOX", "DROP-DOWN", "DROP-DOWN-LIST", "FILL-IN" and "TOGGLE-BOX".
    *         The default value is "FILL-IN". If datatype-exp is not "LOGICAL", specifying
    *         "TOGGLE-BOX" as the widget type generates a run-time error. The widget-type
    *         parameter is ignored in character mode.
    *
    * @return the handle of the new column.
    */
   @Override
   public handle addCalcColumn(character dataType,
                               character format,
                               character initialValue,
                               character label,
                               NumberType position,
                               character widgetType)
   {
      // validate data type
      if (dataType == null || dataType.isUnknown())
      {
         dataType = new character("");
      }

      String dataTypeStr = dataType.toStringMessage();
      ParmType parmType = null;

      if (!dataTypeStr.isEmpty())
      {
         parmType = TempTableBuilder.validateDatatype(dataTypeStr);
      }

      if (parmType == null)
      {
         ErrorManager.recordOrThrowError(9165,
               String.format("Invalid datatype %s given in ADD-CALC-COLUMN", dataTypeStr));
         return new handle();
      }

      // validate format
      String formatStr;

      if (format == null || format.isUnknown() || format.getValue().isEmpty())
      {
         formatStr = TempTableBuilder.getDefaultDataTypeFormat(dataTypeStr);
      }
      else
      {
         formatStr = format.getValue();
         if (!TempTableBuilder.validateFormat(parmType, formatStr))
         {
            ErrorManager.recordOrThrowError(9166,
                  String.format("Invalid format %s given in ADD-CALC-COLUMN", formatStr), false);
            return new handle();
         }
      }

      // form screen value from initial value once - error message may be displayed, but it
      // doesn't prevent addition of the column
      try
      {
         DisplayFormat.instanceOfType(LogicalTerminal.isChui(),
                                      parmType.toString(),
                                      initialValue.getValue(),
                                      formatStr);
      }
      catch (ErrorConditionException e)
      {
         // ignore
      }

      // validate widget type
      BrowseColumnWidgetType columnWidgetType = null;
      String widgetTypeStr = widgetType.toStringMessage().toLowerCase();
      if ("fill-in".equals(widgetTypeStr))
      {
         columnWidgetType = BrowseColumnWidgetType.FILL_IN;
      }
      else if ("toggle-box".equals(widgetTypeStr))
      {
         if (ParmType.LOG.equals(parmType))
         {
            columnWidgetType = TOGGLE_BOX;
         }
      }
      else if ("combo-box".equals(widgetTypeStr))
      {
         if (ParmType.CHAR.equals(parmType) || ParmType.LC.equals(parmType))
         {
            columnWidgetType = BrowseColumnWidgetType.COMBO_BOX;
         }
      }

      if (columnWidgetType == null)
      {
         ErrorManager.recordOrThrowError(4061,
               String.format("**Attribute ADD-CALC-COLUMN for the BROWSE %s has an invalid value",
                     name().toStringMessage()), false);
         return new handle();
      }

      String columnLabel = label != null ? label.getValue() : null;
      BrowseColumnWidget column = addColumnWidget(parmType.toString(),
                                                  formatStr,
                                                  columnLabel,
                                                  position,
                                                  null,
                                                  initialValue,
                                                  columnWidgetType,
                                                  null,
                                                  null,
                                                  true);

      return column.asWidgetHandle();
   }

   /**
    * Implements the ADD-COLUMNS-FROM() widget method.<p>
    * Creates a browse column for each field found in the specified buffer or table.
    *
    * @param  buffHandle
    *         The handle of the buffer associated with query to get data from.
    * @param  exceptList
    *         The optinal comma separated list of the names to be excluded from browse.
    *         Can be empty.
    *
    * @return <code>true</code> if the column was successfully added.
    */
   @Override
   public logical addColumnsFrom(handle buffHandle, character exceptList)
   {
      // check for buffer handle to be valid and query to be set
      if (!buffHandle._isValid() || !checkQueryAssigned())
      {
         new logical(false);
      }
      
      Buffer buff = getBufferForTable(query, buffHandle);
      
      return addColumnsFromInt(buff, exceptList);
   }

   /**
    * Implements the ADD-COLUMNS-FROM() widget method.<p>
    * Creates a browse column for each field found in the specified buffer or table.
    *
    * @param  tableName
    *         The name of the table associated with query to get data from.
    * @param  exceptList
    *         The optinal comma separated list of the names to be excluded from browse.
    *         Can be empty.
    *
    * @return <code>true</code> if the column was successfully added.
    */
   @Override
   public logical addColumnsFrom(character tableName, character exceptList)
   {
      // check for table name validity and query to be set
      if (tableName == null || tableName.isUnknown() || !checkQueryAssigned())
      {
         new logical(false);
      }
      
      Buffer buff = getBufferForTable(query, tableName);
      
      return addColumnsFromInt(buff, exceptList);
   }

   /**
    * Sets the sort-arrow indicator and resets sort-level number for a browse column.
    *
    * @param  columnIndex
    *         1-based index of the column.
    * @param  ascending
    *         <code>true</code> for ascending sorting (arrow up). <code>false</code> for
    *         descending sorting (arrow down).
    *
    * @return <code>true</code> on success.
    */
   @Override
   public logical setSortArrow(int columnIndex, boolean ascending)
   {
      return setSortArrow(new integer(columnIndex), new logical(ascending), new integer());
   }

   /**
    * Sets the sort-arrow indicator and the sort-level number for a browse column.
    *
    * @param  columnIndex
    *         1-based index of the column.
    * @param  ascending
    *         <code>true</code> for ascending sorting (arrow up). <code>false</code> for
    *         descending sorting (arrow down).
    * @param  sortNumber
    *         Sort-level for the browse column. <code>0</code> to hide the sort-level number,
    *         otherwise numbers from <code>1</code> to <code>9</code> are allowed.
    *
    * @return <code>true</code> on success.
    */
   @Override
   public logical setSortArrow(int columnIndex, boolean ascending, int sortNumber)
   {
      return setSortArrow(new integer(columnIndex),
                          new logical(ascending),
                          new integer(sortNumber));
   }

   /**
    * Sets the sort-arrow indicator and resets sort-level number for a browse column.
    *
    * @param  columnIndex
    *         1-based index of the column.
    * @param  ascending
    *         <code>true</code> for ascending sorting (arrow up). <code>false</code> for
    *         descending sorting (arrow down).
    *
    * @return <code>true</code> on success.
    */
   @Override
   public logical setSortArrow(int64 columnIndex, logical ascending)
   {
      return setSortArrow(columnIndex, ascending, new integer());
   }

   /**
    * Sets the sort-arrow indicator and the sort-level number for a browse column.
    *
    * @param  columnIndex
    *         1-based index of the column.
    * @param  ascending
    *         <code>true</code> for ascending sorting (arrow up). <code>false</code> for
    *         descending sorting (arrow down).
    * @param  sortNumber
    *         Sort-level for the browse column. Unknown or <code>0</code> to hide the sort-level
    *         number, otherwise numbers from <code>1</code> to <code>9</code> are allowed.
    *
    * @return <code>true</code> on success.
    */
   @Override
   public logical setSortArrow(int64 columnIndex, logical ascending, int64 sortNumber)
   {
      Integer idx = columnIndex.toJavaIntegerType();
      if (idx == null || idx <= 0 || idx > columns.size())
      {
         ErrorManager.recordOrShowError(4065,
               String.format("**The SET-SORT-ARROW attribute on the BROWSE %s " +
                     "has invalid arguments", name().toStringMessage()),
               false, false);
         return new logical(false);
      }

      BrowseColumnWidget column = columns.get(idx - 1);
      column.setSortNumber(sortNumber);
      if (!column.isSortNumberValid(sortNumber))
      {
         // produce the second error
         ErrorManager.recordOrShowError(4077,
               String.format("**Unable to query attribute SET-SORT-ARROW for BROWSE %s",
                             name().toStringMessage()),
               false, false);
         return new logical(false);
      }
      column.setSortAscending(ascending);
      return new logical(true);
   }

   /**
    * Resets sort-arrow indicators and sort-level numbers for all browse columns.
    *
    * @return Always <code>true</code>.
    */
   @Override
   public logical clearSortArrows()
   {
      for (BrowseColumnWidget column : columns)
      {
         column.setSortAscending(new logical());
         column.setSortNumber(new integer());
      }
      return new logical(true);
   }

   /**
    * Scrolls a browse (if necessary) to bring the currently selected row into view. If the browse
    * supports multiple selections, then SCROLL-TO-CURRENT-ROW() brings the most recently selected
    * row into view.<p>
    * The position of the scrolled row is the first row in the browse viewport, unless the current
    * row is already visible. In this case, the current row remains in the original position.
    *
    * @return <code>true</code> if the row is successfully scrolled into view (or if the scroll is
    *         unnecessary).
    */
   @Override
   public logical scrollToCurrentRow()
   {
      try
      {
         LogicalTerminal.scrollToCurrentRow(this);
         return new logical(true);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, "SCROLL-TO-CURRENT-ROW");
         return new logical(false);
      }
   }

   /**
    * Gets DROP-TARGET attribute.
    * 
    * @return  DROP-TARGET current value.
    */
   @Override
   public logical isDropTarget()
   {
      return new logical(config.dropTarget);
   }
   
   /**
    * Sets DROP-TARGET attribute.
    * 
    * @param   flag
    *          New value for the attribute.
    */
   @Override
   public void setDropTarget(boolean flag)
   {
      if (config.dropTarget != flag)
      {
         config.dropTarget = flag;
         pushWidgetAttr("dropTarget", config.dropTarget);
      }
   }
   
   /**
    * Scrolls a browse (if necessary) to bring the specified selected row into view.
    *
    * @param  selectedRowIndex
    *         Selected row within the browse. A numbered list of selected rows, starting at 1 is
    *         maintained. When the <code>SCROLL-TO-SELECTED-ROW(n)</code> method is encountered,
    *         the list is iterated to find the <i>n</i>th selected row.
    *
    * @return <code>true</code> if the row is successfully scrolled into view (or if the scroll
    *         is unnecessary).
    */
   @Override
   public logical scrollToSelectedRow(int selectedRowIndex)
   {
      return scrollToSelectedRow(new integer(selectedRowIndex));
   }

   /**
    * Scrolls a browse (if necessary) to bring the specified selected row into view.
    *
    * @param  selectedRowIndex
    *         Selected row within the browse. A numbered list of selected rows, starting at 1 is
    *         maintained. When the <code>SCROLL-TO-SELECTED-ROW(n)</code> method is encountered,
    *         the list is iterated to find the <i>n</i>th selected row.
    *
    * @return <code>true</code> if the row is successfully scrolled into view (or if the scroll
    *         is unnecessary).
    */
   @Override
   public logical scrollToSelectedRow(NumberType selectedRowIndex)
   {
      try
      {
         if (selectedRowIndex.isUnknown())
         {
            // caught immediately
            throw new ErrorConditionException(385, "");
         }

         LogicalTerminal.scrollToSelectedRow(this, selectedRowIndex.intValue());
         return new logical(true);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, "SCROLL-TO-SELECTED-ROW");
         return new logical(false);
      }
   }

   /**
    * Get the {@link FwdReport} object associated with this browse.
    *
    * @return the {@link FwdReport} object associated with this browse
    *         or unknown value if reporting is not enabled for this browse.
    */
   @Override
   public handle getReport()
   {
      if (!config.exporting)
      {
         ErrorManager.recordOrShowError(-1,
               "Cannot get BROWSE:REPORT attribute, exporting is not enabled for this browse",
               true, false);
         return new handle();
      }

      if (report == null)
      {
         report = new handle();
         ReportFactory.create(report, this);
      }

      return report;
   }

   /**
    * Get the value of the ENHANCED-SORTING attribute.
    *
    * @return <code>true</code> if enhanced sorting is enabled.
    */
   @Override
   public logical isEnhancedSorting()
   {
      return new logical(config.enhancedSorting);
   }

   /**
    * Get the value of the ENHANCED-FILTERING attribute.
    *
    * @return <code>true</code> if enhanced filtering is enabled.
    */
   @Override
   public logical isEnhancedFiltering()
   {
      return new logical(config.enhancedFiltering);
   }

   /**
    * Sets ENHANCED-SORTING attribute.
    *
    * @param sorting
    *        <code>true</code> to enable enhanced sorting. The attribute cannot be set to
    *        <code>true</code> if enhanced sorting is disabled for the application.
    */
   @Override
   public void setEnhancedSorting(boolean sorting)
   {
      setEnhancedSorting(new logical(sorting));
   }

   /**
    * Sets ENHANCED-SORTING attribute.
    *
    * @param sorting
    *        <code>true</code> to enable enhanced sorting. The attribute cannot be set to
    *        <code>true</code> if enhanced sorting is disabled for the application.
    */
   @Override
   public void setEnhancedSorting(logical sorting)
   {
      if (!sorting.isUnknown())
      {
         config.enhancedSorting = sorting.booleanValue() &&
                               EnhancedBrowseConfigManager.getInstance().isEnhancedSorting();
         pushWidgetAttr("enhancedSorting", config.enhancedSorting);
      }
   }

   /**
    * Sets ENHANCED-FILTERING attribute.
    *
    * @param filtering
    *        <code>true</code> to enable enhanced filtering. The attribute cannot be set to
    *        <code>true</code> if enhanced filtering is disabled for the application.
    */
   @Override
   public void setEnhancedFiltering(boolean filtering)
   {
      setEnhancedFiltering(new logical(filtering));
   }

   /**
    * Sets ENHANCED-FILTERING attribute.
    *
    * @param filtering
    *        <code>true</code> to enable enhanced filtering. The attribute cannot be set to
    *        <code>true</code> if enhanced filtering is disabled for the application.
    */
   @Override
   public void setEnhancedFiltering(logical filtering)
   {
      if (!filtering.isUnknown())
      {
         config.enhancedFiltering = filtering.booleanValue() &&
                                 EnhancedBrowseConfigManager.getInstance().isEnhancedFiltering();
         pushWidgetAttr("enhancedFiltering", config.enhancedFiltering);
      }
   }

   /**
    * Sets NO-ENHANCED-SORTING option.
    *
    * @param noSorting
    *        <code>true</code> to disable enhanced sorting.
    */
   public void setNoEnhancedSorting(boolean noSorting)
   {
      setEnhancedSorting(!noSorting);
   }

   /**
    * Sets NO-ENHANCED-FILTERING option.
    *
    * @param noFiltering
    *        <code>true</code> to disable enhanced filtering.
    */
   public void setNoEnhancedFiltering(boolean noFiltering)
   {
      setEnhancedFiltering(!noFiltering);
   }

   /**
    * Set dynamic sorting for the browse.
    *
    * @param columnOrdinals
    *        Array of the ordinals of the sort columns. Inside the array ordinals are positioned
    *        from the column which defines the most coarse sorting to the most granular.
    * @param ascending
    *        Array of the sorting directions of the sort columns. <code>true</code> for ascending,
    *        <code>false</code> for descending. Inside the array directions are positioned
    *        from the column which defines the most coarse sorting to the most granular.
    */
   public void setDynamicSorting(int[] columnOrdinals, boolean[] ascending)
   {
      if (columnOrdinals == null || columnOrdinals.length == 0)
      {
         setDynamicSorting((FieldReference[]) null, null);
      }
      else
      {
         FieldReference[] refs = new FieldReference[columnOrdinals.length];
         for (int i = 0; i < columnOrdinals.length; i++)
         {
            BrowseColumnWidget column = columns.get(columnOrdinals[i]);
            FieldReference fieldReference = column.getFieldReference();
            if (!column.isEnhancedSorting().booleanValue() || fieldReference == null)
            {
               // safety and security
               throw new IllegalArgumentException("Column is not sortable: " + column.name());
            }

            refs[i] = fieldReference;
         }

         setDynamicSorting(refs, ascending);
      }
   }

   /**
    * Set dynamic filtering for the browse.
    *
    * @param columnOrdinals
    *        Array of the ordinals of the filtered columns.
    * @param values
    *        Array of the filter values for the filtered columns..
    */
   public void setDynamicFiltering(int[] columnOrdinals, BaseDataType[] values)
   {
      if (columnOrdinals == null || columnOrdinals.length == 0)
      {
         setDynamicFiltering((FieldReference[]) null, null, null);
      }
      else
      {
         FieldReference[] refs = new FieldReference[columnOrdinals.length];
         String[] formats = new String[columnOrdinals.length];
         
         for (int i = 0; i < columnOrdinals.length; i++)
         {
            BrowseColumnWidget column = columns.get(columnOrdinals[i]);
            FieldReference fieldReference = column.getFieldReference();
            if (!column.isEnhancedFiltering().booleanValue() || fieldReference == null)
            {
               // safety and security
               throw new IllegalArgumentException("Column is not filterable: " + column.name());
            }
            
            formats[i] =  column.config().format;
            refs[i] = fieldReference;
         }
         
         setDynamicFiltering(refs, values, formats);
      }
   }

   /**
    * Returns string identifying the browse.
    *
    * @return string identifying the browse.
    */
   @Override
   public String toString()
   {
      if (config.ehKeyProcName != null && config.ehKeyBrowseName != null)
      {
         return this.getClass().getName() +
                " id=" + Utils.nonNull(getId(), "<null>") +
               " proc=" + config.ehKeyProcName +
               " name=" + config.ehKeyBrowseName;
      }

      return super.toString();
   }

   /**
    * Get container for getting / setting attributes of the cell in the row for which ROW-DISPLAY
    * event is being triggered.
    *
    * @param  columnIndex
    *         0-based index of the column in the row for which ROW-DISPLAY event is being
    *         triggered. Specifies the cell which attributes we are accessing.
    *
    * @return container for getting / setting attributes of the cell in the row for which
    *         ROW-DISPLAY event is being triggered.
    */
   CellAttributes getRowInDisplayAttributes(int columnIndex)
   {
      if (rowInDisplay == null)
      {
         return null;
      }

      CellAttributes res = rowInDisplay.attr[columnIndex];
      if (res == null)
      {
         res = LogicalTerminal.isChui() ? new ChuiCellAttributes() : new GuiCellAttributes();
         rowInDisplay.attr[columnIndex] = res;
      }

      return res;
   }

   /**
    * Returns <code>true</code> if ROW-DISPLAY trigger is being executed.
    *
    * @return see above.
    */
   boolean isInRowDisplayTrigger()
   {
      return rowInDisplay != null;
   }

   /**
    * If ROW-DISPLAY trigger is being executed, set the screen value of the cell.
    *
    * @param columnIndex
    *        0-based index of the column to which the target cell belongs.
    * @param value
    *        Value to set.
    */
   void setCellScreenValue(int columnIndex, BaseDataType value)
   {
      rowInDisplay.data[columnIndex] = value.duplicate();
   }

   /**
    * Compute the expression kept in the {@link #dynTitle} field and set it
    * as the current browse title. This operation will cancel the old
    * {@link #dynTitle} by setting it to null.
    */
   void updateDynamicTitle()
   {
      if (dynTitle != null)
      {
         character title = dynTitle.execute();
         String    txt   = (title.isUnknown() ? "" : title.toStringMessage());

         setTitle(txt);
      }
   }

   /**
    * Get iterator for browse column widgets.
    *
    * @return iterator for browse column widgets.
    */
   Iterator<BrowseColumnWidget> getBrowseColumns()
   {
      return columns.iterator();
   }

   /**
    * Get visible browse column widgets.
    *
    * @return list of visible browse column widgets.
    */
   public List<BrowseColumnWidget> getVisibleColumns()
   {
      List<BrowseColumnWidget> res = new ArrayList<>();
      for (BrowseColumnWidget column : columns)
      {
         if (column._isVisible())
         {
            res.add(column);
         }
      }
      return res;
   }

   /**
    * Figure out if editing is possible in this browse.
    *
    * @return <code>true</code> if browse is not read-only and there is at least one non-read-only
    *         column.
    */
   boolean calcEditPossible()
   {
      if (isReadOnly().booleanValue())
      {
         return false;
      }

      for (BrowseColumnWidget column : columns)
      {
         if (!column.isReadOnly().booleanValue())
         {
            return true;
         }
      }

      return false;
   }

   /**
    * Create editor widgets for all non-read-only columns.
    */
   void createEditors()
   {
      for (BrowseColumnWidget column : columns)
      {
         BrowseColumnConfig columnConfig = column.config;
         if (!columnConfig.readOnly && column.getEditor() == null)
         {
            GenericWidget editor;
            if (TOGGLE_BOX.equals(columnConfig.widgetType))
            {
               editor = new ToggleBoxWidget(true)
               {
                  @Override
                  public character name()
                  {
                     return column.name();
                  }
               };
            }
            else if (COMBO_BOX.equals(columnConfig.widgetType))
            {
               editor = new ComboBoxWidget(true)
               {
                  @Override
                  public character name()
                  {
                     return column.name();
                  }
               };

               ComboBoxWidget comboBox = (ComboBoxWidget) editor;
               if (columnConfig.items != null && columnConfig.items.length > 0)
               {
                  boolean pairs = columnConfig.pairs != null ? columnConfig.pairs : false;
                  comboBox.setItems(columnConfig.items, pairs);
               }
               comboBox.setHonorFormat(columnConfig.honorFormat);
               comboBox.setInnerLines(columnConfig.innerLines);
               comboBox.setSort(columnConfig.sort);

               if (column.isDropDownList())
               {
                  comboBox.setModeDropDownList(true);
               }
               else
               {
                  comboBox.setModeDropDown(true);
               }
            }
            else
            {
               editor = new FillInWidget(true)
               {
                  @Override
                  public character name()
                  {
                     return column.name();
                  }
               };
            }

            editor.setId(WidgetId.nextID().asInt());
            editor.setDataType(columnConfig.dataType);
            ValidationExpr validationExpr = column.getValidationExpression();
            editor.setValidation(validationExpr, column.getValidationMessage());
            column.config.validatable = validationExpr != null;

            ConfigManager.getInstance().addWidgetConfig(editor.config);
            LogicalTerminal.registerWidget(editor, editor.getId());
            column.setEditor(editor);
         }
      }
   }

   /**
    * Perform actions necessary when a new column was added or visibility of an existing column
    * has changed.
    *
    * @param refreshData
    *        <code>true</code> to refresh row data, <code>false</code> to only update UI.
    */
   void notifyColumnsUpdated(boolean refreshData)
   {
      LogicalTerminal.notifyColumnsUpdated(this, refreshData);
   }

   /**
    * Finish the config processing for this browse.
    * <p>
    * This includes registering the columns' widget IDs with the browse's configuration.
    */
   @Override
   protected void finishConfigProcessing()
   {
      updateColumnIds();
   }

   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    enabledOnly
    *           If <code>true</code>, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   protected logical validateFields(boolean enabledOnly)
   {
      for (BrowseColumnWidget column : columns)
      {
         logical valid = column.validateFields(enabledOnly);
         if (!valid.booleanValue())
         {
            return valid;
         }
      }

      return new logical(true);
   }

   /**
    * Performs initial handling of enhanced configuration, which includes creation of default 4gl
    * state (which matches non-enhanced state) and application the enhanced config saved for this
    * browse (if any).
    */
   private void enhancedInit()
   {
      int colNo = getNumColumns().intValue();
      int exprCounter = 0;
      for (int i = 1; i <= colNo; i++)
      {
         BrowseColumnWidget columnWidget =
               (BrowseColumnWidget) getBrowseColumn(i).unwrapBrowseColumn();
         
         if (columnWidget.config.ehColumnKey == null)
         {
            String columnKey;
            FieldReference fieldReference = columnWidget.getFieldReference();
            if (fieldReference != null)
            {
               columnKey = fieldReference.getFullFieldName();
            }
            else
            {
               columnKey = String.format("expr%2x(%s, %s, %s)",
                     ++exprCounter,
                     new character(columnWidget.config.dataType).toStringExport(),
                     new character(columnWidget.config.format).toStringExport(),
                     new character(columnWidget.config.label).toStringExport());
            }

            columnKey = checkDuplicateColumnKey(columnKey);
            columnWidget.config.ehColumnKey = columnKey;
            ehColumnKeys.add(columnKey);
         }
      }

      config.ehKeyProcName = ProcedureManager.thisProcedure().unwrapProcedure().name().getValue();
      config.ehKeyBrowseName = name().getValue();
      default4glState = new EnhancedBrowseConfig(this);

      EnhancedBrowseConfig cfg =
            EnhancedBrowseConfigManager.getInstance().getEnhancedConfig(default4glState.getKey());
      if (cfg != null)
      {
         applyEnhancedConfig(cfg);
      }

      pushScreenDefinition();
      enhancedInitDone = true;
   }

   /**
    * Create browse column widget and add it to this browse.
    *
    * @param  dataType
    *         Column data type (string representation as returned by {@link ParmType#toString()}).
    * @param  dataFormat
    *         Column data format.
    * @param  columnLabel
    *         Column label.
    * @param  position
    *         1-based position of the browse column (i.e. if pos = 2, the column is the second
    *         column). If the position is <code>null</code> or invalid, the new column is added at
    *         the end of the columns.
    * @param  fieldReference
    *         Reference to the buffer field which backs this column. <code>null</code> if it is a
    *         calc column.
    * @param  initialValue
    *         Initial cell value, if it is a calc column. <code>null</code> if it is a normal
    * @param  widgetType
    *         Widget type: fill-in, toggle-box or combo-box.
    * @param  validateExpression
    *         Validate expression for the column. <code>null</code> if there is no validation for
    *         this widget.
    * @param  validateMessage
    *         Validate message for the column.
    * @param  calcColumn
    *         <code>true</code> if it is a calc column.
    *
    * @return created browse column widget.
    */
   private BrowseColumnWidget addColumnWidget(String dataType,
                                              String dataFormat,
                                              String columnLabel,
                                              NumberType position,
                                              FieldReference fieldReference,
                                              character initialValue,
                                              BrowseColumnWidgetType widgetType,
                                              String validateExpression,
                                              String validateMessage,
                                              boolean calcColumn)
   {
      String columnKey = null;
      if (config.applyEnhancedLayout)
      {
         if (fieldReference != null)
         {
            columnKey = fieldReference.getFullFieldName();
         }
         else
         {
            columnKey = String.format("calc%2x(%s, %s, %s, %s)",
                                      ++calcColumnCounter,
                                      new character(dataType).toStringExport(),
                                      new character(dataFormat).toStringExport(),
                                      initialValue.toStringExport(),
                                      new character(columnLabel).toStringExport());
         }

         columnKey = checkDuplicateColumnKey(columnKey);
      }

      BrowseColumnWidget columnWidget = new BrowseColumnWidget(true, calcColumn);
      columnWidget.setId(WidgetId.nextID().asInt());

      int pos = position == null || position.isUnknown() ? -1 : position.intValue() - 1;
      this.addColumn(columnWidget, pos);
      columnWidget.setDataType(dataType);
      if (dataFormat != null && !dataFormat.isEmpty())
      {
         columnWidget.setFormat(dataFormat);
      }
      columnWidget.setFrame(getFrame());
      columnWidget.setLabel(columnLabel);
      if (fieldReference != null)
      {
         columnWidget.setElement(new Element(fieldReference, columnWidget));

         if (validateExpression != null && !validateExpression.trim().isEmpty())
         {
            JastValidationExpr expr = DynamicValidationHelper.parseValidationExpression(
                  query,
                  (Buffer) fieldReference.getDMO(),
                  fieldReference.getProperty(),
                  validateExpression);
            columnWidget.setValidation(expr, validateMessage);
         }
      }
      else
      {
         BaseDataType value = columnWidget.screenValueToBDT(initialValue, true);
         columnWidget.setElement(new Element(value, columnWidget));
      }

      columnWidget.config.widgetType = widgetType;
      columnWidget.config.ehColumnKey = columnKey;
      if (columnKey != null)
      {
         ehColumnKeys.add(columnKey);
      }
      columnWidget.afterConfigUpdate(columnWidget.config);

      notifyColumnsUpdated(_isVisible() &&
                           getFrame() != null &&
                           getFrame().asWidget().allFramesVisible());
      return columnWidget;
   }

   /**
    * Checks if the given column key already belongs to a column.
    *
    * @param  columnKey
    *         Column key to check.
    *
    * @return <code>columnKey</code> if the given column key DOESN'T belongs to a column (i.e. if
    *         the key is not duplicate). An empty string if the key is duplicate.
    */
   private String checkDuplicateColumnKey(String columnKey)
   {
      if (ehColumnKeys == null)
      {
         ehColumnKeys = new HashSet<>();
      }

      if (ehColumnKeys.contains(columnKey))
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Duplicate column key " + columnKey);
         }
         return "";
      }

      return columnKey;
   }

   /**
    * Fill <code>config.columnIds</code> array with ids of the columns.
    */
   private void updateColumnIds()
   {
      config.columnIds = new int[columns.size()];
      for (int i = 0; i < config.columnIds.length; i++)
      {
         config.columnIds[i] = columns.get(i).getId();
      }
   }

   /**
    * If a character value is displayed at non-character column, change column's data type to
    * "character" and make column right-aligned if the original column format was right-aligned or
    * it was a numeric column.
    *
    * @param column
    *        Config of the target column.
    * @param val
    *        Value to be displayed.
    */
   private void fixDataTypeOverride(BrowseColumnWidget column, BaseDataType val)
   {
      BrowseColumnConfig config = column.config;

      if (!config.dataType.equalsIgnoreCase(val.getTypeName()) &&
            val.getTypeName().equals("CHARACTER"))
      {
         // change column type and format
         // TODO implement override for other data types

         DisplayFormat fmt = Browse.getDisplayFormat(config);

         int len = fmt.getScreenWidth();
         boolean rightAligned = fmt.isRightPositioning() || (fmt instanceof NumberFormat);

         if (len <= 0)
            len = 1;

         config.dataType = "CHARACTER";
         config.format = "x(" + len + ")";

         // reset displayFormat to reflect the format change
         if (config.displayFormat != null)
         {
            config.displayFormat = null;
            Browse.getDisplayFormat(config);
         }

         if (rightAligned)
            config.rightAligned = true;
         
         column.pushWidgetAttr(new String[] {"dataType", "format", "rightAligned"},
                               new Object[] {config.dataType, config.format, config.rightAligned});
      }
   }

   /**
    * Validate WIDTH attribute value.
    *
    * @param  width
    *         WIDTH attribute value to be assigned.
    *
    * @return <code>true</code> if attribute value is valid and can be assigned.
    */
   private boolean validateWidth(NumberType width)
   {
      if (width.isUnknown() || width.doubleValue() <= 0)
      {
         ErrorManager.recordOrShowError(4084,
               String.format("**Attribute WIDTH must be greater than zero on BROWSE widget"),
               false, false);
         return false;
      }

      if (config.realized && LogicalTerminal.isChui())
      {
         errorAlreadyRealized("WIDTH");
         return false;
      }

      return true;
   }

   /**
    * Validate HEIGHT attribute value.
    *
    * @param  height
    *         HEIGHT attribute value to be assigned.
    *
    * @return <code>true</code> if attribute value is valid and can be assigned.
    */
   private boolean validateHeight(NumberType height)
   {
      if (height.isUnknown() || height.doubleValue() <= 0)
      {
         ErrorManager.recordOrShowError(4084,
               String.format("**Attribute HEIGHT must be greater than zero on BROWSE widget"),
               false, false);
         return false;
      }

      if (config.realized && LogicalTerminal.isChui())
      {
         errorAlreadyRealized("HEIGHT");
         return false;
      }

      return true;
   }

   /**
    * Delete the row with the specified index. Query remains repositioned to the next row after
    * the deleted one.
    *
    * @param  rowIndex
    *         0-based index of the deleted row.
    * @param  refreshBrowse
    *         If <code>true</code> then the browse is refreshed.
    *
    * @return <code>true</code> on success.
    */
   private logical deleteRow(int rowIndex, boolean refreshBrowse)
   {
      if (rowIndex >= 0)
      {
         // delete row if it is existing
         reposition(new int64(rowIndex + 1), false);
         logical res = query.deleteResultListEntry(true);
         if (!res.booleanValue())
            return res;
      }

      if (refreshBrowse)
         LogicalTerminal.refreshAfterDeletion(this, rowIndex);
      return new logical(true);
   }

   /**
    * Common reposition code.
    * 
    * @param   row
    *          1-based row to set.
    * @param   fetch
    *          <code>true</code> to auto-fetch the record after reposition,
    *          <code>false</code> to move the cursor only.
    */
   private void reposition(int64 row, boolean fetch)
   {
      if (query == null)
      {
         return;
      }

      boolean fOr = query.isFetchOnReposition(); 
      
      try
      {
         ignoreRepo++;
         
         if (!fetch)
            query.setFetchOnReposition(false);
         
         query.reposition(row);
      }
      finally
      {
         ignoreRepo--;
         
         if (!fetch)
            query.setFetchOnReposition(fOr);
      }
   }

   /**
    * Check if the underlying query is open and the result set contains records. If it doesn't
    * have any records, error message "Browse widget must contain at least one record before
    * <i>methodName</i> method is used" is displayed.
    *
    * @param   methodName
    *          The name of 4GL method that calls this function. Used in the error text.
    *
    * @return  <code>true</code> if the query is open and the result set contains at least one
    *          record.
    */
   private boolean checkNoRecords(String methodName)
   {
      if (query == null || !query.isOpen().getValue() ||
          isEditPossible() ? !LogicalTerminal.hasRows(this) :
                             (query.getNumResults().isUnknown() ||
                              query.getNumResults().intValue() == 0))
      {
         final String msg = String.format("Browse widget must contain at least one " +
               "record before %s method is used", methodName);
         ErrorManager.recordOrShowError(2108, msg, false, false);
         return false;
      }

      return true;
   }

   /**
    * Get number of rows in the browse viewport set either by DOWN or SIZE options.
    *
    * @return  number of rows in the browse viewport.
    */
   private int getBrowsePageRowCount()
   {
      if (pageRowsCount == null)
         pageRowsCount = LogicalTerminal.getBrowsePageRowCount(this);

      return pageRowsCount;
   }

   /**
    * Displays the "Invalid row index specified for browse method <i>methodName</i>" error
    * message.
    *
    * @param methodName
    *        The name of 4GL method that calls this function. Used in the error text.
    */
   private void errorInvalidRowIndex(String methodName)
   {
      final String msg = "Invalid row index specified for browse method " + methodName;
      ErrorManager.recordOrShowError(new int[] { 385 },
                                     new String[] { msg },
                                     false,
                                     false,
                                     true,
                                     false,
                                     false);
   }

   /**
    * Displays the "Invalid use of browse method <i>methodName</i>. There are no selected rows."
    * error message.
    *
    * @param methodName
    *        The name of 4GL method that calls this function. Used in the error text.
    */
   private void errorNoSelectedRows(String methodName)
   {
      final String msg = String.format(
            "Invalid use of browse method %s. There are no selected rows", methodName);
      ErrorManager.recordOrShowError(new int[] { 382 },
                                     new String[] { msg },
                                     false,
                                     false,
                                     true,
                                     false,
                                     false);
   }

   /**
    * Performs full cycle of DELETE-CURRENT-ROW or DELETE-SELECTED-ROW processing.
    *
    * @param  selIndex
    *         0-based position of the row in the list of selected rows. <code>-1</code> for the
    *         last selected row.
    * @param  methodName
    *         4GL method name, used for error messages.
    *
    * @return <code>true</code> of success.
    */
   private logical deleteSingleRow(int selIndex, String methodName)
   {
      if (!checkNoRecords(methodName))
      {
         return new logical(false);
      }

      logical res = new logical(false);
      try
      {
         int targetRow = LogicalTerminal.getSelectedRowIndex(this, selIndex);
         res = deleteRow(targetRow, true);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
      }

      return res;
   }

   /**
    * Displays the appropriate error message for an exception with specific 4GL error code.
    *
    * @param  exc
    *         Exception with specific 4GL error code.
    *
    * @param  methodName
    *         4GL method name, used in the error message.
    */
   private void handleError(ErrorConditionException exc, String methodName)
   {
      int errorCode = exc.getProgressErrorCode();
      if (errorCode == 385)
      {
         errorInvalidRowIndex(methodName);
      }
      else if (errorCode == 382)
      {
         errorNoSelectedRows(methodName);
      }
      else if (errorCode == 2108)
      {
         ErrorManager.recordOrShowError(errorCode,
               String.format("Browse widget must contain at least one record before %s " +
                     "method is used", methodName), false, false);
      }
      else if (errorCode == 6427)
      {
         ErrorManager.recordOrShowError(errorCode,
               String.format("The selected row must be in the viewport when using %s " +
                     "for BROWSE BROWSE %s", //correct
                     methodName, name().toStringMessage()), false, false);
      }
      else
      {
         throw new IllegalStateException("Unknown error code: " + errorCode);
      }
   }

   /**
    * If necessary, init MAX-DATA-GUESS after the query has been (re)opened.
    */
   private void initMaxDataGuess()
   {
      if (maximumKnownRowIndex != null)
         return;

      integer size = query.size();
      maximumKnownRowIndex = size.isUnknown() ? 0 : size.intValue();
      if (query.isNativelyPreselect())
      {
         setMaxDataGuess(maximumKnownRowIndex);
         realDataGuessUsed = true;
         config.maxDataReached = true;
         pushScreenDefinition();
      }
      else if (maximumKnownRowIndex > DATA_GUESS_DEFAULT_VALUE)
      {
         setMaxDataGuess(maximumKnownRowIndex);
         realDataGuessUsed = true;
      }
      else
      {
         setMaxDataGuess(DATA_GUESS_DEFAULT_VALUE);
         realDataGuessUsed = false;
      }
   }

   /**
    * Determine if in-browse editing is possible.
    *
    * @return <code>true</code> if there is at least one editable column.
    */
   private boolean isEditPossible()
   {
      return config.editPossible;
   }

   /**
    * Raises "Unable to do ADD-LIKE-COLUMN because the QUERY is not set" error if the query is
    * not set for this browse.
    *
    * @return <code>true</code> if the query is set.
    */
   private boolean checkQueryAssigned()
   {
      if (query == null)
      {
         ErrorManager.recordOrThrowError(9157,
               "Unable to do ADD-LIKE-COLUMN because the QUERY is not set", false);
         return false;
      }

      return true;
   }

   /**
    * Displays "Unable to set ATTRIBUTE_NAME because the BROWSE widget has been realised."
    * error
    *
    * @param attributeName
    *        Attribute name used in the error message.
    */
   private void errorAlreadyRealized(String attributeName)
   {
      ErrorManager.recordOrShowError(4053,
            String.format("**Unable to set %s because the BROWSE widget has been realised",
                          attributeName),
            false, false);
   }

   /**
    * Supplement the set of rows which represent the upper part of the rows in view with the
    * bottom part (i.e. fetch the set of following rows enough to fill the view).
    *
    * @param  upper
    *         The set of rows to be supplemented, represents the upper part of the rows in view.
    * @param  rowDisplayTriggerId
    *         Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *         trigger for this browse.
    * @param  sb
    *         Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *         ROW-DISPLAY trigger for this browse.
    * @param  displayStrategy
    *         Specifies type of action for which rows are fetched and defines how rows are
    *         iterated / displayed.
    *
    * @return specified set of rows supplemented with the bottom part (at maximum the
    *         supplemented set contains DOWN number of rows).
    */
   private BrowseRow[] addRowsToBottom(BrowseRow[] upper,
                                       int rowDisplayTriggerId,
                                       ScreenBuffer[] sb,
                                       RowDisplayStrategy displayStrategy)
   {
      int rowsToFetch = config.down - upper.length;
      if (rowsToFetch > 0)
      {
         // still not enough rows to fill the view
         int divisionRow = upper[upper.length - 1].rowIndex;
         BrowseRow[] lowerPart = getRows(divisionRow + 1,
                                         rowsToFetch,
                                         true,
                                         rowDisplayTriggerId,
                                         sb,
                                         displayStrategy);
         if (lowerPart.length > 0)
         {
            BrowseRow[] newResult = new BrowseRow[upper.length + lowerPart.length];
            System.arraycopy(upper, 0, newResult, 0, upper.length);
            System.arraycopy(lowerPart, 0, newResult, upper.length, lowerPart.length);
            return newResult;
         }
      }

      return upper;
   }

   /**
    * Returns <code>true</code> if the target row is in view.
    *
    * @param  viewportTopRow
    *         0-based index (in the browse model) of the top row in view.
    * @param  viewportBottomRow
    *         0-based index (in the browse model) of the bottom row in view.
    * @param  targetRow
    *         0-based index (in the browse model) of the target row.
    *
    * @return <code>true</code> if the target row is in view.
    */
   private boolean isRowInView(int viewportTopRow, int viewportBottomRow, int targetRow)
   {
      return viewportTopRow >= 0 &&
             viewportBottomRow >= 0 &&
             targetRow >= viewportTopRow &&
             targetRow <= viewportBottomRow;
   }

   /**
    * Returns <code>true</code> if the target row is on the first "page" of the result set.
    *
    * @param  targetRow
    *         Target row to check.
    *
    * @return <code>true</code> if the target row is on the first "page" of the result set.
    */
   private boolean isRowOnFirstPage(int targetRow)
   {
      return targetRow < config.down; // TODO take row gaps into account?
   }

   /**
    * Raise the error with the specific code related to row height.
    *
    * @param code
    *        4GL error code.
    * @param pixels
    *        <code>true</code> if the error is for ROW-HEIGHT-PIXELS attribute, <code>false</code>
    *        if the error is for ROW-HEIGHT(-CHARS) attribute.
    */
   private void rowHeightError(int code, boolean pixels)
   {
      String attributeName = pixels ? "ROW-HEIGHT-PIXELS" : "ROW-HEIGHT";

      switch (code)
      {
         case 6422:
            ErrorManager.recordOrShowError(6422, String.format(ERROR_TEXT_6422,
                                                               attributeName,
                                                               name().toStringMessage(),
                                                               MIN_BROWSE_ROWS),
                                           false);
            break;

         case 3131:
            ErrorManager.recordOrShowError(3131,
                  String.format("Unable to set attribute %s in widget %s of type %s",
                        attributeName, name().toStringMessage(), type()), false, false);
            break;

         case 4083:
            ErrorManager.recordOrShowError(4083, String.format(
                  "Unable to assign UNKNOWN value to attribute %s on %s %s", attributeName,
                  type(), name().toStringMessage()), false);
            break;

         case 4060:
            ErrorManager.recordOrShowError(4060,
                  String.format("Attribute %s for the %s %s has an invalid value",
                        attributeName, type(), name().toStringMessage()), false);
            break;
      }
   }

   /**
    * Get array of the configurations of the columns.
    *
    * @return see above.
    */
   private BrowseColumnConfig[] getColumnConfigs()
   {
      int size = columns.size();
      BrowseColumnConfig[] columnConfigs = new BrowseColumnConfig[size];
      for (int colIdx = 0; colIdx < size; colIdx++)
      {
         columnConfigs[colIdx] = columns.get(colIdx).config();
      }
      return columnConfigs;
   }

   /**
    * Get field reference(s) as name array for given query and buffer pair.
    *
    * @param  buffer
    *         Handle to the buffer associated with browse query.
    *
    * @return reference to the target fields name array if <code>buffer</code> is valid and
    *         references a buffer which participates in the query, empty array otherwise.
    */
   private ArrayList<String> getFieldsForBuffer(Buffer buffer)
   {
      ArrayList<String> retList = new ArrayList<String>();
      
      // table name does not associated with query
      if (buffer == null)
      {
         return retList;
      }
      
      // getting the field names from buffer 
      int numFieldsCount = buffer.numFields().intValue();
      for (int j = 1; j <= numFieldsCount; j++)
      {
         BufferField bufferField = (BufferField) buffer.bufferField(j).getResource();
         retList.add(((Nameable) bufferField).name().getValue());
      }
      
      return retList;
   }
   
   /**
    * Get buffer associated with the given table name and query specified.
    *
    * @param  queryToCheck
    *         The query which backs the browse was called for.
    * @param  tableName
    *         The table name to search within buffers of the given query.
    *
    * @return Buffer instance associated with the table name if found, <code>null</code>
    *         otherwise.
    */
   private Buffer getBufferForTable(P2JQuery queryToCheck, character tableName)
   {
      // simple check for valid table name
      String tableNameStr = tableName.toStringMessage();
      // the table name is empty, return error
      if (tableNameStr.isEmpty())
      {
         return null;
      }
      
      Buffer buffer = null;
      
      int bufferCount = queryToCheck.numBuffers().intValue();
      for (int i = 1; i <= bufferCount; i++)
      {
         Buffer bufferNext = queryToCheck.bufferHandle(i).unwrapBuffer();
         // table name associated with next buffer in current query
         String bufferNameNext = bufferNext.name().getValue();
       
         if (bufferNameNext.equals(tableNameStr))
         {
            buffer = bufferNext;
            break;
         }
      }
      
      return buffer;
   }

   /**
    * Get buffer associated with the given buffer handle and query specified.
    *
    * @param  queryToCheck
    *         The query which backs the browse was called for.
    * @param  buffHandle
    *         The buffer handle to search within buffers of the given query.
    *
    * @return Buffer instance associated with the buffer handle if found, <code>null</code>
    *         otherwise.
    */
   private Buffer getBufferForTable(P2JQuery queryToCheck, handle buffHandle)
   {
      Buffer buffer = buffHandle.unwrapBuffer();
      
      boolean match = false;
      int bufferCount = queryToCheck.numBuffers().intValue();
      for (int i = 1; i <= bufferCount; i++)
      {
         Buffer bufferNext = queryToCheck.bufferHandle(i).unwrapBuffer();
         if (bufferNext.equals(buffer))
         {
            match = true;
            break;
         }
      }

      // not fount the buffer in current query 
      if (!match)
      {
         return null;
      }
      else
      {
         return buffer;
      }
   }
   
   /**
    * Implements the ADD-COLUMNS-FROM() widget method.  Internal commom worker.<p>
    * Creates a browse column for each field found in the specified buffer or table.
    *
    * @param  srcBuffer
    *         The buffer of the table associated with query to get data from.
    * @param  exceptList
    *         The optinal comma separated list of the names to be excluded from browse.
    *         Can be empty.
    *
    * @return <code>true</code> if the column was successfully added.
    */
   private logical addColumnsFromInt(Buffer srcBuffer, character exceptList)
   {
      // table buffer is not associated with query
      if (srcBuffer == null)
      {
         return new logical(false);
      }
      
      boolean retCode = false;
      
      // gets field list from the given buffer
      ArrayList<String> fieldsToAdd = getFieldsForBuffer(srcBuffer);
      if (fieldsToAdd.size() > 0)
      {
         // getting here means we are ready to add the fields
         retCode = true;
         // reconstruct fields exception list
         String excpt[] = exceptList == null || exceptList.isUnknown()
                             ? null : exceptList.toStringMessage().split(",");
         // remove exept fields from list to be added
         if (excpt != null && excpt.length > 0)
         {
            for (int i = 0; i < excpt.length; i++)
            {
               // remove exception entry if existed from fields array to add
               fieldsToAdd.remove(excpt[i]);
            }
         }
         // initialize added columns map if required
         if(colsAddedFrom == null)
         {
            colsAddedFrom = new HashMap<String, ArrayList<String>>();
         }
         // now lets check if the fields already exists as some of the browse column
         // we keep track of only fields added by ADD-COLUMNS-FROM method
         // get the table name
         String tableName = srcBuffer.name().getValue();
         // and list of the previously added fields
         if (colsAddedFrom.containsKey(tableName))
         {
            for (String addedName : colsAddedFrom.get(tableName))
            {
               if(addedName != null && !addedName.isEmpty())
               {
                  // remove alrady added field from fields array to add
                  fieldsToAdd.remove(addedName);
               }
            }
         }
         else
         {
            // create empty storage for newly added fields
            // this is the part of the strange 4GL behavior described above
            colsAddedFrom.put(tableName, new ArrayList<String>());
         }
         // if at this time we still have smething to add - do it
         for (String fieldToAdd : fieldsToAdd)
         {
            String fullFieldName = new StringBuilder().append(tableName)
                                                      .append(".")
                                                      .append(fieldToAdd).toString();
                                                      
            handle hcol = addLikeColumn(fullFieldName);
            // something went wrong, return false
            if (hcol._isValid())
            {
               // need to update added fields registry
               // this is the part of the strange 4GL behavior described above
               colsAddedFrom.get(tableName).add(fieldToAdd);
            }
            else
            {
               retCode = false;
               break;
            }
         }
      }
      
      return new logical(retCode);
   }

   /**
    * Matches buffers of two queries. If the buffers match, the method returns {@code true}. Two buffers
    * match if the underlying fields of both buffers equal in data types.
    *
    * @param   query1
    *          A query.
    * @param   query2
    *          A query.
    *
    * @return  see above.
    */
   private boolean matchBufferFields(QueryWrapper query1, QueryWrapper query2)
   {
      if (query1 == null || query2 == null)
      {
         return false;
      }

      int count1 = query1.numBuffers().intValue();
      int count2 = query2.numBuffers().intValue();

      if (count1 != count2)
      {
         return false;
      }

      for (int i = 1; i <= count1; i++)
      {
         Buffer buf1 = (Buffer) query1.bufferHandle(i).getResource();
         Buffer buf2 = (Buffer) query2.bufferHandle(i).getResource();

         int fcount1 = buf1.numFields().intValue();
         int fcount2 = buf2.numFields().intValue();

         if (fcount1 != fcount2)
         {
            return false;
         }

         for (int j = 1; j <= fcount1; j++)
         {
            BufferField f1 = (BufferField) buf1.bufferField(j).getResource();
            BufferField f2 = (BufferField) buf2.bufferField(j).getResource();

            if (f1.extent().intValue() != f2.extent().intValue())
            {
               return false;
            }

            if (!Objects.equals(f1.getDataType(), f2.getDataType()))
            {
               return false;
            }
         }
      }

      return true;
   }
   
   /**
    * Pending query reopening state while REFRESHABLE is off.
    */
   private enum PendingOpenState
   {
      /** Query wasn't reopen. */
      NO_REOPEN,

      /**
       * Query was reopened, but initial reposition for the browsed query wasn't performed.
       * See {@link com.goldencode.p2j.persist.AbstractQuery#open()}.
       */
      PRE_REOPEN,

      /**
       * Query was reopened, and initial reposition for the browsed query was performed.
       * See {@link com.goldencode.p2j.persist.AbstractQuery#open()}.
       */
      REOPEN,

      /**
       * Query was reopened, but initial reposition for the browsed query wasn't performed, and
       * there was a reposition <i>before</i> query reopening.
       * See {@link com.goldencode.p2j.persist.AbstractQuery#open()}.
       */
      PRE_REPOSITIONED_AND_REOPEN,

      /**
       * Query was reopened, and initial reposition for the browsed query was performed, and
       * there was a reposition <i>before</i> query reopening.
       * See {@link com.goldencode.p2j.persist.AbstractQuery#open()}.
       */
      REPOSITIONED_AND_REOPEN
   }
}
